'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from typing import Optional, Dict, Any, List, Set, Iterable

import polars as pl

from pm4py.objects.log.util.dataframe_utils import Parameters
from pm4py.util import constants, exec_utils, pandas_utils
from pm4py.util import xes_constants


def _dedupe_preserve_order(values: Iterable[str]) -> List[str]:
    seen: Set[str] = set()
    out: List[str] = []
    for v in values:
        if v not in seen:
            seen.add(v)
            out.append(v)
    return out


def _sanitize_feature_name(
    prefix: str, value: Any, used_names: Optional[Set[str]] = None
) -> str:
    sanitized = str(value).encode("ascii", errors="ignore").decode("ascii")
    if not sanitized:
        sanitized = "value"
    base_name = f"{prefix}_{sanitized}"

    if used_names is None:
        return base_name

    candidate = base_name
    suffix = 1
    while candidate in used_names:
        candidate = f"{base_name}__{suffix}"
        suffix += 1
    used_names.add(candidate)
    return candidate


def _scalar_from_lazy(lf: pl.LazyFrame, expr: pl.Expr) -> Any:
    result = lf.select(expr.alias("__scalar")).collect()
    if result.height == 0 or result.width == 0:
        return None
    return result.to_series(0)[0]


def _lazy_schema(lf: pl.LazyFrame) -> pl.Schema:
    return lf.collect_schema()


def _lazy_columns(lf: pl.LazyFrame) -> List[str]:
    return _lazy_schema(lf).names()


def _drop_if_present(lf: pl.LazyFrame, cols: Iterable[str]) -> pl.LazyFrame:
    existing = set(_lazy_columns(lf))
    to_drop = [c for c in cols if c in existing]
    return lf.drop(to_drop) if to_drop else lf


def _unique_internal_name(existing: Set[str], base: str) -> str:
    if base not in existing:
        return base
    i = 1
    while f"{base}__{i}" in existing:
        i += 1
    return f"{base}__{i}"


def _is_numeric_dtype(dtype: pl.DataType) -> bool:
    dtype_str = str(dtype).lower()
    if any(token in dtype_str for token in ("int", "uint", "float")):
        return True
    if dtype_str.startswith("decimal"):
        return True
    if dtype_str.startswith("duration"):
        return True
    if dtype_str == "boolean":
        return True
    return False


def _is_string_dtype(dtype: pl.DataType) -> bool:
    dtype_str = str(dtype).lower()
    return dtype_str in {"utf8", "string"} or dtype_str.startswith("categorical")


def automatic_feature_selection_df(
    df: pl.LazyFrame, parameters: Optional[Dict[Any, Any]] = None
) -> pl.LazyFrame:
    """Selects useful features from a Polars lazyframe for ML purposes."""
    if parameters is None:
        parameters = {}

    schema = _lazy_schema(df)
    available_columns = set(schema.names())

    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )

    default_mandatory = available_columns.intersection(
        {case_id_key, activity_key, timestamp_key}
    )
    mandatory_attributes = exec_utils.get_param_value(
        Parameters.MANDATORY_ATTRIBUTES,
        parameters,
        default_mandatory,
    )
    mandatory_set = set(mandatory_attributes)

    min_different_occ_str_attr = exec_utils.get_param_value(
        Parameters.MIN_DIFFERENT_OCC_STR_ATTR, parameters, 5
    )
    max_different_occ_str_attr = exec_utils.get_param_value(
        Parameters.MAX_DIFFERENT_OCC_STR_ATTR, parameters, 50
    )
    consider_all_attributes = exec_utils.get_param_value(
        Parameters.CONSIDER_ALL_ATTRIBUTES, parameters, True
    )

    other_attributes_to_retain = set()

    total_cases = _scalar_from_lazy(df, pl.col(case_id_key).n_unique())
    total_cases = int(total_cases or 0)

    for col, dtype in schema.items():
        if col == case_id_key:
            continue

        cases_with_value = _scalar_from_lazy(
            df.filter(pl.col(col).is_not_null()),
            pl.col(case_id_key).n_unique(),
        )
        cases_with_value = int(cases_with_value or 0)

        if cases_with_value != total_cases and not consider_all_attributes:
            continue

        if _is_numeric_dtype(dtype):
            other_attributes_to_retain.add(col)
        elif _is_string_dtype(dtype):
            unique_val_count = _scalar_from_lazy(
                df.filter(pl.col(col).is_not_null()),
                pl.col(col).n_unique(),
            )
            unique_val_count = int(unique_val_count or 0)
            if (
                min_different_occ_str_attr
                <= unique_val_count
                <= max_different_occ_str_attr
            ):
                other_attributes_to_retain.add(col)

    attributes_to_retain = mandatory_set.union(other_attributes_to_retain)
    selected_columns = [
        col_name for col_name in schema.names() if col_name in attributes_to_retain
    ]

    return df.select(selected_columns)


def select_number_column(
    df: pl.LazyFrame,
    fea_df: pl.LazyFrame,
    col: str,
    case_id_key: str = constants.CASE_CONCEPT_NAME,
) -> pl.LazyFrame:
    """Adds a numeric column to the feature lazyframe.

    Notes on column duplication:
      * If `fea_df` already contained `col` (e.g., repeated calls / duplicate inputs),
        Polars would create `col_right` during the join. We explicitly drop any prior
        versions first to keep the output schema stable.
      * We also ensure the internal row-number column does not collide with user data.
    """
    fea_df = _drop_if_present(fea_df, [col, f"{col}_right"])

    df_cols = set(_lazy_columns(df))
    row_nr_col = _unique_internal_name(df_cols, "__row_nr")

    df_numeric = (
        df.with_row_count(row_nr_col)
        .select(pl.col(case_id_key), pl.col(col), pl.col(row_nr_col))
        .drop_nulls(subset=[col])
        .group_by(case_id_key)
        .agg(pl.col(col).sort_by(pl.col(row_nr_col)).last().alias(col))
    )

    return (
        fea_df.join(df_numeric, on=case_id_key, how="left", coalesce=True)
        .with_columns(pl.col(col).cast(pl.Float32))
    )


def _collect_categorical_values(
    df: pl.LazyFrame, columns: List[str]
) -> Dict[str, List[Any]]:
    """Collects formatted unique values for the provided categorical columns."""
    collected: Dict[str, List[Any]] = {}
    for col in columns:
        unique_values = (
            df.select(pl.col(col))
            .drop_nulls(subset=[col])
            .unique()
            .collect()
            .get_column(col)
            .to_list()
        )
        formatted = [
            value
            for value in pandas_utils.format_unique(unique_values)
            if value is not None
        ]
        if formatted:
            collected[col] = formatted

    return collected


def _select_string_columns(
    df: pl.LazyFrame,
    fea_df: pl.LazyFrame,
    columns: List[str],
    case_id_key: str,
    count_occurrences: bool,
) -> pl.LazyFrame:
    """Adds one-hot or count encoded columns for the provided categorical attributes.

    This function is designed to be idempotent: running it multiple times with the same
    inputs will overwrite/recompute the same generated feature columns instead of
    creating suffixed duplicates (e.g., `...__1`, `..._right`).
    """
    if not columns:
        return fea_df

    df_schema = _lazy_schema(df)
    available = set(df_schema.names())

    clean_columns = [
        c
        for c in _dedupe_preserve_order(columns)
        if c != case_id_key and c in available
    ]
    if not clean_columns:
        return fea_df

    unique_values_map = _collect_categorical_values(df, clean_columns)
    if not unique_values_map:
        return fea_df

    existing_cols: Set[str] = set(_lazy_columns(fea_df))
    used_names: Set[str] = set(existing_cols)

    agg_exprs: List[pl.Expr] = []
    fill_exprs: List[pl.Expr] = []
    cols_to_drop: Set[str] = set()

    for column, unique_values in unique_values_map.items():
        for value in unique_values:
            # Deterministic base name (no dependency on existing columns).
            base_name = _sanitize_feature_name(column, value)

            # If the feature column already exists, we recompute it (drop first),
            # avoiding `*_right` duplicates from joins.
            if base_name in existing_cols:
                cols_to_drop.add(base_name)
                used_names.discard(base_name)
            if f"{base_name}_right" in existing_cols:
                cols_to_drop.add(f"{base_name}_right")
                used_names.discard(f"{base_name}_right")

            # Ensure uniqueness against the remaining schema + other new features.
            column_name = base_name
            suffix = 1
            while column_name in used_names:
                column_name = f"{base_name}__{suffix}"
                suffix += 1
            used_names.add(column_name)

            comparison = pl.col(column).eq(value)
            if count_occurrences:
                agg_exprs.append(comparison.cast(pl.Int64).sum().alias(column_name))
            else:
                agg_exprs.append(comparison.cast(pl.Int8).max().alias(column_name))

            fill_exprs.append(pl.col(column_name).cast(pl.Float32))

    if cols_to_drop:
        fea_df = _drop_if_present(fea_df, cols_to_drop)

    feature_chunk = (
        df.select([pl.col(case_id_key)] + [pl.col(c) for c in unique_values_map.keys()])
        .group_by(case_id_key)
        .agg(agg_exprs)
        # Materialize all encoded columns in one go to minimize separate joins.
        .with_columns(fill_exprs)
    )

    return fea_df.join(feature_chunk, on=case_id_key, how="left", coalesce=True)


def select_string_column(
    df: pl.LazyFrame,
    fea_df: pl.LazyFrame,
    col: str,
    case_id_key: str = constants.CASE_CONCEPT_NAME,
    count_occurrences: bool = False,
) -> pl.LazyFrame:
    """Adds one-hot or count encoded columns for a categorical attribute."""
    return _select_string_columns(
        df,
        fea_df,
        [col],
        case_id_key=case_id_key,
        count_occurrences=count_occurrences,
    )


def select_string_columns(
    df: pl.LazyFrame,
    fea_df: pl.LazyFrame,
    columns: List[str],
    case_id_key: str = constants.CASE_CONCEPT_NAME,
    count_occurrences: bool = False,
) -> pl.LazyFrame:
    """Adds one-hot or count encoded columns for the provided categorical attributes."""
    return _select_string_columns(
        df,
        fea_df,
        columns,
        case_id_key=case_id_key,
        count_occurrences=count_occurrences,
    )


def get_features_df(
    df: pl.LazyFrame,
    list_columns: List[str],
    parameters: Optional[Dict[Any, Any]] = None,
) -> pl.LazyFrame:
    """Performs automatic feature extraction on a Polars LazyFrame."""
    if parameters is None:
        parameters = {}

    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )
    add_case_identifier_column = exec_utils.get_param_value(
        Parameters.ADD_CASE_IDENTIFIER_COLUMN, parameters, False
    )
    count_occurrences = exec_utils.get_param_value(
        Parameters.COUNT_OCCURRENCES, parameters, False
    )

    # Avoid duplicate work and join-induced `*_right` columns when the
    # input list contains duplicates.
    list_columns = _dedupe_preserve_order(list_columns)

    fea_df = df.select(pl.col(case_id_key)).unique().sort(case_id_key)

    schema = _lazy_schema(df)
    numeric_columns: List[str] = []
    string_columns: List[str] = []

    for col in list_columns:
        if col == case_id_key or col not in schema:
            continue
        dtype = schema[col]
        if _is_numeric_dtype(dtype):
            numeric_columns.append(col)
        elif _is_string_dtype(dtype):
            string_columns.append(col)

    for col in numeric_columns:
        fea_df = select_number_column(df, fea_df, col, case_id_key=case_id_key)

    fea_df = select_string_columns(
        df,
        fea_df,
        string_columns,
        case_id_key=case_id_key,
        count_occurrences=count_occurrences,
    )

    fea_df = fea_df.sort(case_id_key)
    if not add_case_identifier_column:
        fea_df = fea_df.drop(case_id_key)

    return fea_df


def automatic_feature_extraction_df(
    df: pl.LazyFrame, parameters: Optional[Dict[Any, Any]] = None
) -> pl.LazyFrame:
    """Wrapper that performs automatic feature extraction on a Polars lazyframe."""
    if parameters is None:
        parameters = {}

    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )

    fea_sel_df = automatic_feature_selection_df(df, parameters=parameters)
    columns = set(_lazy_columns(fea_sel_df))

    columns.discard(case_id_key)
    columns.discard(timestamp_key)

    return get_features_df(fea_sel_df, list(columns), parameters=parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from pm4py.util import exec_utils, constants, xes_constants
from pm4py.objects.log.obj import EventLog, EventStream
import pandas as pd
from typing import Union, Dict, Optional, Any, Tuple, List
from pm4py.objects.conversion.log import converter as log_converter


class Parameters(Enum):
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    ENABLE_PADDING = "enable_padding"
    PAD_SIZE = "pad_size"


def apply(
    log: Union[EventLog, EventStream, pd.DataFrame],
    parameters: Optional[Dict[Any, Any]] = None,
) -> Tuple[List[List[int]], List[str]]:
    """
    Returns a list of lists (one for every case of the log) containing the difference between the timestamp of the current event
    and the timestamp of the next event of the case (an automatic padding option is also available).
    For the last event of the case, the difference is defaulted to 0.

    Parameters
    ---------------
    log
        Event log
    parameters
        Parameters of the algorithm, including:
        - Parameters.TIMESTAMP_KEY => the attribute of the log to be used as timestamp
        - Parameters.ENABLE_PADDING => enables the padding (the length of cases is normalized)
        - Parameters.PAD_SIZE => the size of the padding

    Returns
    ---------------
    target
        The aforementioned list
    classes
        Dummy list (of classes)
    """
    if parameters is None:
        parameters = {}

    log = log_converter.apply(
        log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters
    )
    max_case_length = max([len(x) for x in log])

    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    enable_padding = exec_utils.get_param_value(
        Parameters.ENABLE_PADDING, parameters, False
    )
    pad_size = exec_utils.get_param_value(
        Parameters.PAD_SIZE, parameters, max_case_length
    )

    target = []
    for trace in log:
        target.append([])
        for i in range(len(trace)):
            curr_time = trace[i][timestamp_key].timestamp()
            next_time = (
                trace[i + 1][timestamp_key].timestamp()
                if i < len(trace) - 1
                else curr_time
            )
            target[-1].append(float(next_time - curr_time))
        if enable_padding:
            while len(target[-1]) < pad_size:
                target[-1].append(0.0)

    return target, ["@@next_time"]



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import heapq
from pm4py.objects.petri_net.utils.align_utils import  levenshtein, discountedEditDistance
from pm4py.objects.petri_net.utils.petri_utils import  decorate_places_preset_trans,  decorate_transitions_prepostset
from pm4py.objects.petri_net.utils import align_utils as utils
from pm4py.util import exec_utils
from copy import copy
from enum import Enum
from pm4py.statistics.variants.log import get as variants_module

'''
This algorithm computes discounted multi-alignment. 
More details in Boltenhagen's thesis 
Author: Author: Boltenhagen Mathilde, Thomas Chatain, Josep Carmona
Date: Jan. 2021
'''

class Parameters(Enum):
    MARKING_LIMIT = "marking_limit"
    EXPONENT = "exponent"


def apply(log, petri_net, ini, fin, parameters=None):
    """
    This function is a preliminary function for __search of the dijkstra_exponential_heuristic for multi-alignment.
    The function gets the parameters and launches the algorithm on the variants (all traces aren't usefull for MA, see Boltenhagen's thesis).
    """

    # get variants
    variants_idxs = variants_module.get_variants_from_log_trace_idx(log, parameters=parameters)
    variants = []
    for index_variant, var in enumerate(variants_idxs):
        variants.append(var.split(","))

    # get the number of times we can reach the same marking, very usefull for concurrency and loops
    marking_limit =  exec_utils.get_param_value(Parameters.MARKING_LIMIT, parameters, None)
    if marking_limit is None:
        marking_limit=100000

    # the discounted parameter
    exponent = exec_utils.get_param_value(Parameters.EXPONENT, parameters, None)
    if exponent is None:
        exponent=2

    return __search(petri_net, ini, fin, variants,  marking_limit=marking_limit, exponent=exponent)


def __search(net, ini, fin, variants, exponent=2, marking_limit=1000):
    '''
    This function is the A* algorithm for multi-alignments.
    '''

    decorate_transitions_prepostset(net)
    decorate_places_preset_trans(net)

    mymemory={}
    closed = {}

    ini_state = utils.DijkstraSearchTupleForAntiAndMulti(0, ini, [])
    open_set = [ini_state]
    heapq.heapify(open_set)
    visited = 0
    queued = 0
    traversed = 0
    best, sizeAA = None, None

    distanceTime = 0
    trans_empty_preset = set(t for t in net.transitions if len(t.in_arcs) == 0)

    def costfunction(t, curr_ma, withFrac=True):
        '''
        Compute the maximal distance of the current multi-alignment + t and the variants.
        :param t: transition that we want to add in the current multi-alignment
        :param curr_ma: current prefix of multi-alignment
        :param withFrac: the fraction that prevents the best suffice, this parameter is false at the end of the algorithm.
        '''
        if t is None :
            str_aa = [a.label for a in curr_ma]
        elif len(curr_ma)==0 :
            str_aa = [t.label]
        else :
            str_aa = [a.label for a in curr_ma] + [t.label]

        all= []
        for v in variants:
            if str(v+str_aa) not in mymemory.keys() or not withFrac:
                size_of_alignment, distance = discountedEditDistance(str_aa,v, exponent)
                if withFrac:
                    mymemory[str(v+str_aa)] = distance - (exponent**(-len(str_aa)+1)-exponent**(-(len(str_aa)+len(v))))/(exponent-1)
                else :
                    mymemory[str(v+str_aa)] = distance
            all.append(mymemory[str(v+str_aa)])
        return max(all)

    # ----------------------------
    # a* algorithm starts here
    while not len(open_set) == 0:
        curr = heapq.heappop(open_set)

        if best and  best.g < curr.g:
            continue

        if curr.m == fin:
            curr.g = costfunction(None, curr.r, withFrac=False)
            # in case there only one run
            if not best and (len(heapq.nsmallest(1,open_set))==0 or heapq.nsmallest(1,open_set)[0].g > curr.g):
                return  {'multi-alignment': curr.r, 'cost': curr.g, 'visited_states': visited, 'queued_states': queued,
                         'traversed_arcs': traversed, "max_distance_to_log":getMaxDist(curr.r,variants)}

            # other runs might be interesting
            elif not best or (best and  best.g > curr.g):
                best = curr
            continue

        closed[curr.m]= 1 if (curr.m) not in closed.keys() else  closed[curr.m]+1

        visited += 1

        possible_enabling_transitions = copy(trans_empty_preset)
        for p in curr.m:
            for t in p.ass_trans:
                possible_enabling_transitions.add(t)

        enabled_trans = [t for t in possible_enabling_transitions if t.sub_marking <= curr.m]
        trans_to_visit_with_cost = [(t, costfunction(t,curr.r, withFrac=True)) for t in enabled_trans if t is not None]

        for t, cost in trans_to_visit_with_cost:
            traversed += 1
            new_marking = utils.add_markings(curr.m, t.add_marking)

            if ((new_marking) in closed.keys() and closed[new_marking]>marking_limit) or cost==curr.g:
                continue
            queued += 1
            tp = utils.DijkstraSearchTupleForAntiAndMulti(cost, new_marking, curr.r + [t])
            heapq.heappush(open_set, tp)

    return  {'multi-alignment': best.r, 'cost': best.g, 'visited_states': visited, 'queued_states': queued,
             'traversed_arcs': traversed, "max_distance_to_log":getMaxDist(best.r,variants)}

def getMaxDist(run, variants):
    '''
    We give the maximal levenshtein edit distance to the variants
    '''
    run = [a.label for a in run]
    all= []
    for v in variants:
        all.append(levenshtein(run,v))
    return max(all)


'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

import heapq
import time
from collections.abc import Iterator

from pm4py.objects.log import obj as log_implementation
from pm4py.util.xes_constants import DEFAULT_NAME_KEY
from pm4py.objects.petri_net.utils.synchronous_product import (
    construct_cost_aware,
    construct,
)
from pm4py.objects.petri_net.utils.petri_utils import (
    construct_trace_net_cost_aware,
    decorate_places_preset_trans,
    decorate_transitions_prepostset,
)
from pm4py.objects.petri_net.utils import align_utils as utils
from pm4py.util import exec_utils
from copy import copy
from enum import Enum
import sys
from pm4py.util.constants import PARAMETER_CONSTANT_ACTIVITY_KEY
from typing import Optional, Dict, Any, Union, List as TList
from pm4py.objects.log.obj import Trace
from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.util import typing


class Parameters(Enum):
    PARAM_TRACE_COST_FUNCTION = "trace_cost_function"
    PARAM_MODEL_COST_FUNCTION = "model_cost_function"
    PARAM_SYNC_COST_FUNCTION = "sync_cost_function"
    PARAM_ALIGNMENT_RESULT_IS_SYNC_PROD_AWARE = "ret_tuple_as_trans_desc"
    PARAM_TRACE_NET_COSTS = "trace_net_costs"
    TRACE_NET_CONSTR_FUNCTION = "trace_net_constr_function"
    TRACE_NET_COST_AWARE_CONSTR_FUNCTION = (
        "trace_net_cost_aware_constr_function"
    )
    PARAM_MAX_ALIGN_TIME_TRACE = "max_align_time_trace"
    PARAM_MAX_ALIGN_TIME = "max_align_time"
    PARAMETER_VARIANT_DELIMITER = "variant_delimiter"
    ACTIVITY_KEY = PARAMETER_CONSTANT_ACTIVITY_KEY
    VARIANTS_IDX = "variants_idx"


def get_best_worst_cost(
    petri_net, initial_marking, final_marking, parameters=None
):
    """
    Gets the best worst cost of an alignment

    Parameters
    -----------
    petri_net
        Petri net
    initial_marking
        Initial marking
    final_marking
        Final marking

    Returns
    -----------
    best_worst_cost
        Best worst cost of alignment
    """
    if parameters is None:
        parameters = {}
    trace = log_implementation.Trace()

    best_worst = next(
        apply(
            trace,
            petri_net,
            initial_marking,
            final_marking,
            parameters=parameters,
        )
    )

    if best_worst is not None:
        return best_worst["cost"]

    return None


def apply(
    trace: Trace,
    petri_net: PetriNet,
    initial_marking: Marking,
    final_marking: Marking,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Iterator[typing.AlignmentResult]:
    """
    Performs the basic alignment search, given a trace and a net.

    Parameters
    ----------
    trace: :class:`list` input trace, assumed to be a list of events (i.e. the code will use the activity key
    to get the attributes)
    petri_net: :class:`pm4py.objects.petri.net.PetriNet` the Petri net to use in the alignment
    initial_marking: :class:`pm4py.objects.petri.net.Marking` initial marking in the Petri net
    final_marking: :class:`pm4py.objects.petri.net.Marking` final marking in the Petri net
    parameters: :class:`dict` (optional) dictionary containing one of the following:
        Parameters.PARAM_TRACE_COST_FUNCTION: :class:`list` (parameter) mapping of each index of the trace to a positive cost value
        Parameters.PARAM_MODEL_COST_FUNCTION: :class:`dict` (parameter) mapping of each transition in the model to corresponding
        model cost
        Parameters.PARAM_SYNC_COST_FUNCTION: :class:`dict` (parameter) mapping of each transition in the model to corresponding
        synchronous costs
        Parameters.ACTIVITY_KEY: :class:`str` (parameter) key to use to identify the activity described by the events

    Returns
    -------
    dictionary: `dict` with keys **alignment**, **cost**, **visited_states**, **queued_states** and **traversed_arcs**
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, DEFAULT_NAME_KEY
    )
    trace_cost_function = exec_utils.get_param_value(
        Parameters.PARAM_TRACE_COST_FUNCTION, parameters, None
    )
    model_cost_function = exec_utils.get_param_value(
        Parameters.PARAM_MODEL_COST_FUNCTION, parameters, None
    )
    trace_net_constr_function = exec_utils.get_param_value(
        Parameters.TRACE_NET_CONSTR_FUNCTION, parameters, None
    )
    trace_net_cost_aware_constr_function = exec_utils.get_param_value(
        Parameters.TRACE_NET_COST_AWARE_CONSTR_FUNCTION,
        parameters,
        construct_trace_net_cost_aware,
    )

    if trace_cost_function is None:
        trace_cost_function = list(
            map(lambda e: utils.STD_MODEL_LOG_MOVE_COST, trace)
        )
        parameters[Parameters.PARAM_TRACE_COST_FUNCTION] = trace_cost_function

    if model_cost_function is None:
        # reset variables value
        model_cost_function = dict()
        sync_cost_function = dict()
        for t in petri_net.transitions:
            if t.label is not None:
                model_cost_function[t] = utils.STD_MODEL_LOG_MOVE_COST
                sync_cost_function[t] = utils.STD_SYNC_COST
            else:
                model_cost_function[t] = utils.STD_TAU_COST
        parameters[Parameters.PARAM_MODEL_COST_FUNCTION] = model_cost_function
        parameters[Parameters.PARAM_SYNC_COST_FUNCTION] = sync_cost_function

    if trace_net_constr_function is not None:
        # keep the possibility to pass TRACE_NET_CONSTR_FUNCTION in this old
        # version
        trace_net, trace_im, trace_fm = trace_net_constr_function(
            trace, activity_key=activity_key
        )
    else:
        (
            trace_net,
            trace_im,
            trace_fm,
            parameters[Parameters.PARAM_TRACE_NET_COSTS],
        ) = trace_net_cost_aware_constr_function(
            trace, trace_cost_function, activity_key=activity_key
        )

    yield from apply_trace_net(
        petri_net,
        initial_marking,
        final_marking,
        trace_net,
        trace_im,
        trace_fm,
        parameters,
    )


def apply_trace_net(
    petri_net,
    initial_marking,
    final_marking,
    trace_net,
    trace_im,
    trace_fm,
    parameters=None,
):
    """
    Performs the basic alignment search, given a trace net and a net.

    Parameters
    ----------
    trace: :class:`list` input trace, assumed to be a list of events (i.e. the code will use the activity key
    to get the attributes)
    petri_net: :class:`pm4py.objects.petri.net.PetriNet` the Petri net to use in the alignment
    initial_marking: :class:`pm4py.objects.petri.net.Marking` initial marking in the Petri net
    final_marking: :class:`pm4py.objects.petri.net.Marking` final marking in the Petri net
    parameters: :class:`dict` (optional) dictionary containing one of the following:
        Parameters.PARAM_TRACE_COST_FUNCTION: :class:`list` (parameter) mapping of each index of the trace to a positive cost value
        Parameters.PARAM_MODEL_COST_FUNCTION: :class:`dict` (parameter) mapping of each transition in the model to corresponding
        model cost
        Parameters.PARAM_SYNC_COST_FUNCTION: :class:`dict` (parameter) mapping of each transition in the model to corresponding
        synchronous costs
        Parameters.ACTIVITY_KEY: :class:`str` (parameter) key to use to identify the activity described by the events
        Parameters.PARAM_TRACE_NET_COSTS: :class:`dict` (parameter) mapping between transitions and costs

    Returns
    -------
    dictionary: `dict` with keys **alignment**, **cost**, **visited_states**, **queued_states** and **traversed_arcs**
    """
    if parameters is None:
        parameters = {}

    ret_tuple_as_trans_desc = exec_utils.get_param_value(
        Parameters.PARAM_ALIGNMENT_RESULT_IS_SYNC_PROD_AWARE, parameters, False
    )

    trace_cost_function = exec_utils.get_param_value(
        Parameters.PARAM_TRACE_COST_FUNCTION, parameters, None
    )
    model_cost_function = exec_utils.get_param_value(
        Parameters.PARAM_MODEL_COST_FUNCTION, parameters, None
    )
    sync_cost_function = exec_utils.get_param_value(
        Parameters.PARAM_SYNC_COST_FUNCTION, parameters, None
    )
    trace_net_costs = exec_utils.get_param_value(
        Parameters.PARAM_TRACE_NET_COSTS, parameters, None
    )

    if (
        trace_cost_function is None
        or model_cost_function is None
        or sync_cost_function is None
    ):
        sync_prod, sync_initial_marking, sync_final_marking = construct(
            trace_net,
            trace_im,
            trace_fm,
            petri_net,
            initial_marking,
            final_marking,
            utils.SKIP,
        )
        cost_function = utils.construct_standard_cost_function(
            sync_prod, utils.SKIP
        )
    else:
        revised_sync = dict()
        for t_trace in trace_net.transitions:
            for t_model in petri_net.transitions:
                if t_trace.label == t_model.label:
                    revised_sync[(t_trace, t_model)] = sync_cost_function[
                        t_model
                    ]

        sync_prod, sync_initial_marking, sync_final_marking, cost_function = (
            construct_cost_aware(
                trace_net,
                trace_im,
                trace_fm,
                petri_net,
                initial_marking,
                final_marking,
                utils.SKIP,
                trace_net_costs,
                model_cost_function,
                revised_sync,
            )
        )

    max_align_time_trace = exec_utils.get_param_value(
        Parameters.PARAM_MAX_ALIGN_TIME_TRACE, parameters, sys.maxsize
    )

    yield from apply_sync_prod(
        sync_prod,
        sync_initial_marking,
        sync_final_marking,
        cost_function,
        utils.SKIP,
        ret_tuple_as_trans_desc=ret_tuple_as_trans_desc,
        max_align_time_trace=max_align_time_trace,
    )


def apply_sync_prod(
    sync_prod,
    initial_marking,
    final_marking,
    cost_function,
    skip,
    ret_tuple_as_trans_desc=False,
    max_align_time_trace=sys.maxsize,
):
    yield from __search(
        sync_prod,
        initial_marking,
        final_marking,
        cost_function,
        skip,
        ret_tuple_as_trans_desc=ret_tuple_as_trans_desc,
        max_align_time_trace=max_align_time_trace,
    )


def __search(
    sync_net,
    ini,
    fin,
    cost_function,
    skip,
    ret_tuple_as_trans_desc=False,
    max_align_time_trace=sys.maxsize,
):
    start_time = time.time()

    # gviz = vizapply( net=sync_net, initial_marking=ini, final_marking=fin, variant=vizvariant.WO_DECORATION )
    # vizview(gviz)

    decorate_transitions_prepostset(sync_net)
    decorate_places_preset_trans(sync_net)

    closed = set()
    hub: dict[Marking, list[utils.DijkstraSearchTuple]] = dict()
    hub_dist: dict[Marking, int] = dict()
    # Given a marking, holds many tuples of (search tuples that reached it,
    # distance) in two arrays

    ini_state = utils.DijkstraSearchTuple(0, ini, None, None, 0)
    open_set = [ini_state]
    hub[ini_state.m] = [ini_state]
    hub_dist[ini_state.m] = 0
    heapq.heapify(open_set)
    visited = 0
    queued = 0
    traversed = 0

    trans_empty_preset = set(
        t for t in sync_net.transitions if len(t.in_arcs) == 0
    )

    while not len(open_set) == 0:
        if (time.time() - start_time) > max_align_time_trace:
            return None

        curr = heapq.heappop(open_set)
        current_marking = curr.m

        # We don't use closed, but check visited path for any loops.
        # if check_cycle_in_searchtuple(curr) :
        #     continue

        # Prevent recursion on already-closed markings (once is enough)
        # Required because multiple paths can add a marking before it get
        # visited.
        already_closed = current_marking in closed
        if already_closed:
            continue

        # print("MARKING[v={},q={},t={},h={}]: {}   {}".format(visited, queued, traversed, len(open_set), repr(current_marking), hash(current_marking)))
        # print(repr_searchtuple(curr))

        closed.add(current_marking)
        visited += 1

        # If we try to expand the `fin` node , then we can be sure that no better (more optimal)
        # paths can exist, since the heap would have visited them before (and there is always a cost
        # to reach the final node).
        if current_marking == fin:
            for alignment in reconstruct_alignment_generator(
                hub, curr, ret_tuple_as_trans_desc=ret_tuple_as_trans_desc
            ):
                yield {
                    "alignment": alignment,
                    "cost": curr.g,
                    "visited_states": visited,
                    "queued_states": queued,
                    "traversed_arcs": traversed,
                    "lp_solved": False,
                }
            return  # No transition may originate from the fin node.

        enabled_trans = copy(trans_empty_preset)
        for p in current_marking:
            for t in p.ass_trans:
                if t.sub_marking <= current_marking:
                    enabled_trans.add(t)

        trans_to_visit_with_cost = [
            (t, cost_function[t])
            for t in enabled_trans
            if not (
                t is not None
                and utils.__is_log_move(t, skip)
                and utils.__is_model_move(t, skip)
            )
        ]

        for t, cost in trans_to_visit_with_cost:
            traversed += 1
            new_marking = utils.add_markings(
                current_marking, t.add_marking
            )  # Counter operations.
            tp = utils.DijkstraSearchTuple(
                g=curr.g + cost, m=new_marking, p=curr, t=t, l=curr.l + 1
            )

            # We don't use closed, but check visited path for any loops.
            # if check_cycle_in_searchtuple(tp) :
            #     continue
            if new_marking in closed:
                continue

            # Each marking contains all paths that reached it. Used later for reconstruction.
            # We also make sure that each hub node only contains paths with the
            # minimal distance.
            this_cost = curr.g + cost
            if new_marking not in hub:
                hub[new_marking] = []
                hub_dist[new_marking] = this_cost
            min_cost = hub_dist[new_marking]
            if this_cost < min_cost:
                hub[new_marking] = [tp]
                hub_dist[new_marking] = this_cost
            elif this_cost == min_cost:
                hub[new_marking] = hub[new_marking] + [tp]
            else:
                pass  # ignore if this one is bigger

            queued += 1

            heapq.heappush(open_set, tp)


def check_cycle_in_searchtuple(st: utils.DijkstraSearchTuple) -> bool:
    rep = set()
    while st is not None:
        if st.m in rep:
            return True
        rep.add(st.m)
        st = st.p
    return False


def repr_searchtuple(st: utils.DijkstraSearchTuple) -> str:
    out = []
    while st is not None:
        out.append("{}+{}".format(repr(st.m), hash(st.m)))
        st = st.p
    return " -> ".join(reversed(out))


def rec_hub(hub, curr_m) -> TList[TList[utils.DijkstraSearchTuple]]:

    out = []
    for curr_st in hub[curr_m]:
        if curr_st.p is not None:
            prefixes = rec_hub(hub, curr_st.p.m)
            for pref in prefixes:
                out.append(pref + [curr_st])
        else:
            out.append([curr_st])

    return out


def reconstruct_alignment_generator(
    hub, state: utils.DijkstraSearchTuple, ret_tuple_as_trans_desc=False
):
    # print("Reconstruction at: {}".format(state.m))
    paths = rec_hub(hub, state.m)
    # for hk,hv in sorted(hub.items()) :
    #     print("{:>65} : {}  x  {}".format(
    #         str(hk),
    #         len(hv) if hv is not None else 0,
    #         " , ".join([ ( str(hhv.p.m) if hhv.p is not None else "None" ) for hhv in hv ]),
    #     ))
    for path in paths:
        # print("New alignment:")
        alignment = list()
        for pst in path:
            # print("   {}  -  {}".format(pst.m,pst.t.label if pst.t is not None else "NO-T"))
            if pst.t is not None:
                if ret_tuple_as_trans_desc:
                    alignment.append((pst.t.name, pst.t.label))
                else:
                    alignment.append(pst.t.label)

        yield alignment



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from itertools import combinations
from typing import Optional, Tuple, List, TypeVar, Generic, Dict, Any

from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL
from pm4py.algo.discovery.powl.inductive.fall_through.empty_traces import (
    POWLEmptyTracesUVCL,
)
from pm4py.algo.discovery.inductive.variants.abc import InductiveMinerFramework

from pm4py.algo.discovery.powl.inductive.base_case.factory import (
    BaseCaseFactory,
)
from pm4py.algo.discovery.powl.inductive.cuts.factory import CutFactory
from pm4py.algo.discovery.powl.inductive.fall_through.factory import (
    FallThroughFactory,
)
from pm4py.algo.discovery.powl.inductive.utils.filtering import (
    FILTERING_TYPE,
    FilteringType,
    filter_most_frequent_variants,
    FILTERING_THRESHOLD,
    filter_most_frequent_variants_with_decreasing_factor,
    DEFAULT_FILTERING_TYPE,
)
from pm4py.algo.discovery.powl.inductive.variants.powl_discovery_varaints import (
    POWLDiscoveryVariant, )

from pm4py.objects.powl.obj import (
    POWL,
    StrictPartialOrder,
    Sequence,
    OperatorPOWL,
)

T = TypeVar("T", bound=IMDataStructureUVCL)


class IMBasePOWL(Generic[T], InductiveMinerFramework[T]):

    def instance(self) -> POWLDiscoveryVariant:
        return POWLDiscoveryVariant.TREE

    def apply(
        self,
        obj: IMDataStructureUVCL,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> POWL:
        if FILTERING_TYPE not in parameters.keys():
            filtering_type = DEFAULT_FILTERING_TYPE
        else:
            filtering_type = parameters[FILTERING_TYPE]
            if filtering_type not in FilteringType:
                raise KeyError(
                    "Invalid FILTERING_TYPE: " + str(filtering_type)
                )

        empty_traces = POWLEmptyTracesUVCL.apply(obj, parameters)
        if empty_traces is not None:
            return self._recurse(empty_traces[0], empty_traces[1], parameters)

        powl = self.apply_base_cases(obj, parameters)
        if powl is not None:
            return powl

        cut = self.find_cut(obj, parameters)
        if cut is not None:
            powl = self._recurse(cut[0], cut[1], parameters=parameters)

        if powl is not None:
            return powl
        else:
            if filtering_type is FilteringType.DYNAMIC:
                filtered_log = filter_most_frequent_variants(
                    obj.data_structure
                )
                if len(filtered_log.data_structure) > 0:
                    return self.apply(filtered_log, parameters=parameters)

            elif filtering_type is FilteringType.DECREASING_FACTOR:
                if FILTERING_THRESHOLD in parameters.keys():
                    t = parameters[FILTERING_THRESHOLD]
                    if isinstance(t, float) and 0 <= t < 1:
                        if t > 0:
                            filtered_log = filter_most_frequent_variants_with_decreasing_factor(
                                obj.data_structure, decreasing_factor=t)
                            if (
                                0
                                < len(filtered_log.data_structure)
                                < len(obj.data_structure)
                            ):
                                return self.apply(
                                    filtered_log, parameters=parameters
                                )
                    else:
                        raise KeyError("Invalid filtering threshold!")
            else:
                raise KeyError("Invalid filtering type!")

            ft = self.fall_through(obj, parameters)
            return self._recurse(ft[0], ft[1], parameters=parameters)

    def apply_base_cases(
        self, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[POWL]:
        return BaseCaseFactory.apply_base_cases(obj, parameters=parameters)

    def find_cut(
        self, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[Tuple[POWL, List[T]]]:
        return CutFactory.find_cut(obj, parameters=parameters)

    def fall_through(
        self, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Tuple[POWL, List[T]]:
        return FallThroughFactory.fall_through(
            obj, self._pool, self._manager, parameters=parameters
        )

    def _recurse(
        self,
        powl: POWL,
        objs: List[T],
        parameters: Optional[Dict[str, Any]] = None,
    ):
        children = [self.apply(obj, parameters=parameters) for obj in objs]
        if isinstance(powl, StrictPartialOrder):
            powl_new = StrictPartialOrder(children)
            for i, j in combinations(range(len(powl.children)), 2):
                if powl.order.is_edge_id(i, j):
                    powl_new.order.add_edge(children[i], children[j])
                elif powl.order.is_edge_id(j, i):
                    powl_new.order.add_edge(children[j], children[i])
            return powl_new
        else:
            new_powl = OperatorPOWL(operator=powl.operator, children=children)
            return new_powl



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from collections import Counter
from pm4py.objects.log.obj import EventLog, Event, Trace
from pm4py.util import xes_constants as xes_util
import heapq
from pm4py.objects.petri_net.utils.petri_utils import (
    decorate_places_preset_trans,
    decorate_transitions_prepostset,
)
from pm4py.objects.petri_net.utils import align_utils as utils
from pm4py.objects.petri_net.utils.incidence_matrix import construct
from pm4py.util import constants, pandas_utils
import pandas as pd


def __search(sync_net, ini, fin, stop, cost_function, skip):
    decorate_transitions_prepostset(sync_net)
    decorate_places_preset_trans(sync_net)

    incidence_matrix = construct(sync_net)
    ini_vec, fin_vec, cost_vec = utils.__vectorize_initial_final_cost(
        incidence_matrix, ini, fin, cost_function
    )

    closed = set()

    ini_state = utils.SearchTuple(0, 0, 0, ini, None, None, None, True)
    open_set = [ini_state]
    heapq.heapify(open_set)
    visited = 0
    queued = 0
    traversed = 0

    # return all the prefix markings of the optimal alignments as set
    ret_markings = None
    # keep track of the optimal cost of an alignment (to trim search when
    # needed)
    optimal_cost = None

    while not len(open_set) == 0:
        curr = heapq.heappop(open_set)

        current_marking = curr.m

        # trim alignments when we already reached an optimal alignment and the
        # current cost is greater than the optimal cost
        if optimal_cost is not None and curr.f > optimal_cost:
            break

        already_closed = current_marking in closed
        if already_closed:
            continue

        if stop <= current_marking:
            # add the current marking to the set
            # of returned markings
            if ret_markings is None:
                ret_markings = set()
            ret_markings.add(current_marking)
            # close the marking
            closed.add(current_marking)
            # set the optimal cost
            optimal_cost = curr.f

            continue

        closed.add(current_marking)
        visited += 1

        enabled_trans = set()
        for p in current_marking:
            for t in p.ass_trans:
                if t.sub_marking <= current_marking:
                    enabled_trans.add(t)

        trans_to_visit_with_cost = [
            (t, cost_function[t])
            for t in enabled_trans
            if not (
                t is None
                or utils.__is_log_move(t, skip)
                or (utils.__is_model_move(t, skip) and not t.label[1] is None)
            )
        ]

        for t, cost in trans_to_visit_with_cost:
            traversed += 1
            new_marking = utils.add_markings(current_marking, t.add_marking)

            if new_marking in closed:
                continue
            g = curr.g + cost

            queued += 1
            new_f = g

            tp = utils.SearchTuple(
                new_f, g, 0, new_marking, curr, t, None, True
            )
            heapq.heappush(open_set, tp)

    return ret_markings


def get_log_prefixes(
    log,
    activity_key=xes_util.DEFAULT_NAME_KEY,
    case_id_key=constants.CASE_CONCEPT_NAME,
):
    """
    Get log prefixes

    Parameters
    ----------
    log
        Trace log
    activity_key
        Activity key (must be provided if different from concept:name)
    """
    prefixes = {}
    prefix_count = Counter()

    if pandas_utils.check_is_pandas_dataframe(log):
        traces = pandas_utils.get_traces(log, case_id_key, activity_key)
    else:
        traces = [tuple(x[activity_key] for x in trace) for trace in log]

    for trace in traces:
        for i in range(1, len(trace)):
            prefix = constants.DEFAULT_VARIANT_SEP.join(trace[0:i])
            next_activity = trace[i]
            if prefix not in prefixes:
                prefixes[prefix] = set()
            prefixes[prefix].add(next_activity)
            prefix_count[prefix] += 1

    return prefixes, prefix_count


def form_fake_log(prefixes_keys, activity_key=xes_util.DEFAULT_NAME_KEY):
    """
    Form fake log for replay (putting each prefix as separate trace to align)

    Parameters
    ----------
    prefixes_keys
        Keys of the prefixes (to form a log with a given order)
    activity_key
        Activity key (must be provided if different from concept:name)
    """
    fake_log = EventLog()
    for prefix in prefixes_keys:
        trace = Trace()
        prefix_activities = prefix.split(constants.DEFAULT_VARIANT_SEP)
        for activity in prefix_activities:
            event = Event()
            event[activity_key] = activity
            trace.append(event)
        fake_log.append(trace)
    return fake_log



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util.constants import CASE_CONCEPT_NAME, PARAMETER_CONSTANT_CASEID_KEY, PARAMETER_CONSTANT_TIMESTAMP_KEY, PARAMETER_CONSTANT_ACTIVITY_KEY
from pm4py.util.xes_constants import DEFAULT_TIMESTAMP_KEY, DEFAULT_NAME_KEY
from enum import Enum
from pm4py.util import exec_utils
from typing import Optional, Dict, Any, Union
import polars as pl


class Parameters(Enum):
    CASE_ID_KEY = PARAMETER_CONSTANT_CASEID_KEY
    ACTIVITY_KEY = PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = PARAMETER_CONSTANT_TIMESTAMP_KEY
    FILTER_TYPE = "filter_type"


def apply(
    df: pl.LazyFrame,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.LazyFrame:
    """
    Groups the events of the same case happening at the same timestamp,
    providing option to keep the first event of each group, keep the last event of each group, create an event
    having as activity the concatenation of the activities happening in the group

    Parameters
    -----------
    df
        LazyFrame
    parameters
        Parameters of the algorithm, including:
        - Parameters.CASE_ID_KEY => the case identifier to be used
        - Parameters.ACTIVITY_KEY => the attribute to be used as activity
        - Parameters.TIMESTAMP_KEY => the attribute to be used as timestamp
        - Parameters.FILTER_TYPE => the type of filter to be applied:
            first => keeps the first event of each group
            last => keeps the last event of each group
            concat => creates an event having as activity the concatenation of the activities happening in the group

    Returns
    -----------
    df
        Filtered LazyFrame
    """
    if parameters is None:
        parameters = {}

    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, DEFAULT_TIMESTAMP_KEY
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, DEFAULT_NAME_KEY
    )
    filter_type = exec_utils.get_param_value(
        Parameters.FILTER_TYPE, parameters, "first"
    )

    if filter_type == "first":
        # Keep first event of each group
        df = df.group_by([case_id_key, timestamp_key], maintain_order=True).first()
    elif filter_type == "last":
        # Keep last event of each group
        df = df.group_by([case_id_key, timestamp_key], maintain_order=True).last()
    elif filter_type == "concat":
        # Get all columns except activity for the first event, then concatenate activities
        # First, get the grouped dataframe
        grouped = df.group_by([case_id_key, timestamp_key], maintain_order=True)
        
        # Get all columns from first event
        first_df = grouped.first()
        
        # Get concatenated activities
        concat_activities = grouped.agg(
            pl.col(activity_key).sort().str.concat(" & ").alias(activity_key)
        )
        
        # Update the activity column in first_df with concatenated values
        df = first_df.drop(activity_key).join(
            concat_activities,
            on=[case_id_key, timestamp_key],
            how="left"
        )

    return df


'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util.vis_utils import (
    human_readable_stat,
    get_arc_penwidth,
    get_trans_freq_color,
)


def get_shortest_paths_from_trans(
    original_trans,
    trans,
    spaths,
    visited_arcs,
    visited_transitions,
    added_elements,
    rec_depth,
):
    """
    Get shortest paths from a given transition

    Parameters
    --------------
    original_trans
        Original transition
    trans
        Current considered transition
    spaths
        Map of shortest paths
    visited_arcs
        Set of visited arcs
    visited_transitions
        Set of visited transitions
    added_elements
        Elements to add recursively
    rec_depth
        Recursion depth

    Returns
    -------------
    spaths
        Map of shortest paths
    visited_arcs
        Set of visited arcs
    added_elements
        Elements to add recursively
    """
    for out_arc in trans.out_arcs:
        if out_arc not in visited_arcs:
            visited_arcs.add(out_arc)
            target_place = out_arc.target
            for place_out_arc in target_place.out_arcs:
                if place_out_arc not in visited_arcs:
                    visited_arcs.add(place_out_arc)
                    target_trans = place_out_arc.target
                    if target_trans not in visited_transitions:
                        visited_transitions.add(target_trans)
                        if target_trans.label:
                            el1 = (
                                (original_trans.name, target_trans.name),
                                0,
                                rec_depth,
                            )
                            if out_arc not in spaths:
                                spaths[out_arc] = set()
                            spaths[out_arc].add(el1)
                            added_elements.add(el1)
                            el2 = (
                                (original_trans.name, target_trans.name),
                                1,
                                rec_depth,
                            )
                            if place_out_arc not in spaths:
                                spaths[place_out_arc] = set()
                            spaths[place_out_arc].add(el2)
                            added_elements.add(el2)
                        else:
                            (
                                spaths,
                                visited_arcs,
                                visited_transitions,
                                added_elements,
                            ) = get_shortest_paths_from_trans(
                                original_trans,
                                target_trans,
                                spaths,
                                visited_arcs,
                                visited_transitions,
                                added_elements,
                                rec_depth + 1,
                            )
                            for element in added_elements:
                                new_element = list(element)
                                if new_element[1] == 0:
                                    new_element[1] = 2
                                    if out_arc not in spaths:
                                        spaths[out_arc] = set()
                                    spaths[out_arc].add(tuple(new_element))
                                if new_element[1] == 1:
                                    new_element[1] = 3
                                    if place_out_arc not in spaths:
                                        spaths[place_out_arc] = set()
                                    spaths[place_out_arc].add(
                                        tuple(new_element)
                                    )
    return spaths, visited_arcs, visited_transitions, added_elements


def get_shortest_paths(net, enable_extension=False):
    """
    Gets shortest paths between visible transitions in a Petri net

    Parameters
    -----------
    net
        Petri net
    enable_extension
        Enable decoration of more arcs, in a risky way, when needed

    Returns
    -----------
    spaths
        Shortest paths
    """
    spaths = {}
    for trans in net.transitions:
        if trans.label:
            visited_arcs = set()
            visited_transitions = set()
            added_elements = set()
            spaths, visited_arcs, visited_transitions, added_elements = (
                get_shortest_paths_from_trans(
                    trans,
                    trans,
                    spaths,
                    visited_arcs,
                    visited_transitions,
                    added_elements,
                    0,
                )
            )
    spaths_keys = list(spaths.keys())
    for edge in spaths_keys:
        list_zeros = [el for el in spaths[edge] if el[1] == 0]
        list_ones = [el for el in spaths[edge] if el[1] == 1]
        if list_zeros:
            spaths[edge] = {x for x in spaths[edge] if x[1] == 0}
            min_dist = min([x[2] for x in spaths[edge]])
            possible_targets = set(
                [x[0] for x in spaths[edge] if x[2] == min_dist]
            )
            spaths[edge] = set()
            for target in possible_targets:
                spaths[edge].add((target, 0, min_dist))
        elif list_ones:
            spaths[edge] = {x for x in spaths[edge] if x[1] == 1}
            min_dist = min([x[2] for x in spaths[edge]])
            possible_targets = set(
                [x[0] for x in spaths[edge] if x[2] == min_dist]
            )
            spaths[edge] = set()
            for target in possible_targets:
                spaths[edge].add((target, 1, min_dist))
        else:
            unique_targets = set([x[0] for x in spaths[edge]])
            if len(unique_targets) == 1:
                spaths[edge] = set()
                spaths[edge].add((list(unique_targets)[0], 2, 0))
            else:
                if enable_extension:
                    min_dist = min([x[2] for x in spaths[edge]])
                    possible_targets = set(
                        [x[0] for x in spaths[edge] if x[2] == min_dist]
                    )
                    spaths[edge] = set()
                    for target in possible_targets:
                        spaths[edge].add((target, 2, min_dist))
                else:
                    del spaths[edge]

    return spaths


def get_decorations_from_dfg_spaths_acticount(
    net,
    dfg,
    spaths,
    activities_count,
    variant="frequency",
    aggregation_measure=None,
):
    """
    Get decorations from Petrinet without doing any replay
    but based on DFG measures, shortest paths and activities count.
    The variant could be 'frequency' or 'performance'.
    Aggregation measure could also be specified

    Parameters
    -----------
    net
        Petri net
    dfg
        Directly-Follows graph
    spaths
        Shortest paths between visible transitions in the Petri net
    activities_count
        Count of activities in the Petri net
    variant
        Describe how to decorate the Petri net (could be frequency or performance)
    aggregation_measure
        Specifies the aggregation measure
    stat_locale
        Dict to locale the stat strings

    Returns
    -----------
    decorations
        Decorations to use for the Petri net
    """
    from statistics import mean, median, stdev

    decorations_single_contrib = {}
    decorations_single_contrib_trans = {}
    decorations_int = {}
    decorations = {}
    if aggregation_measure is None:
        if "frequency" in variant:
            aggregation_measure = "sum"
        elif "performance" in variant:
            aggregation_measure = "mean"
    for arc in spaths:
        for couple in spaths[arc]:
            dfg_key = couple[0]
            if dfg_key in dfg:
                if arc not in decorations_single_contrib:
                    decorations_single_contrib[arc] = []
                decorations_single_contrib[arc].append(dfg[dfg_key])
                if dfg_key[1] not in decorations_single_contrib_trans:
                    decorations_single_contrib_trans[dfg_key[1]] = {}
                decorations_single_contrib_trans[dfg_key[1]][dfg_key[0]] = dfg[
                    dfg_key
                ]
    for arc in decorations_single_contrib:
        decorations_value = None
        if aggregation_measure == "sum":
            decorations_value = sum(decorations_single_contrib[arc])
        elif aggregation_measure == "mean":
            decorations_value = mean(decorations_single_contrib[arc])
        elif aggregation_measure == "median":
            decorations_value = median(decorations_single_contrib[arc])
        elif aggregation_measure == "stdev":
            decorations_value = stdev(decorations_single_contrib[arc])
        elif aggregation_measure == "min":
            decorations_value = min(decorations_single_contrib[arc])
        elif aggregation_measure == "max":
            decorations_value = max(decorations_single_contrib[arc])
        if decorations_value is not None:
            decorations_int[arc] = decorations_value

    if decorations_int:
        arcs_min_value = min(list(decorations_int.values()))
        arcs_max_value = max(list(decorations_int.values()))
        for arc in decorations_int:
            if "performance" in variant:
                arc_label = human_readable_stat(decorations_int[arc], stat_locale)
            else:
                arc_label = str(decorations_int[arc])
            decorations[arc] = {
                "label": arc_label,
                "penwidth": str(
                    get_arc_penwidth(
                        decorations_int[arc], arcs_min_value, arcs_max_value
                    )
                ),
            }
        trans_map = {}
        for trans in net.transitions:
            if trans.label:
                trans_map[trans.label] = trans
        if "frequency" in variant:
            act_min_value = min(list(activities_count.values()))
            act_max_value = max(list(activities_count.values()))
            for act in activities_count:
                if act in trans_map:
                    trans = trans_map[act]
                    color = get_trans_freq_color(
                        activities_count[act], act_min_value, act_max_value
                    )
                    label = act + " (" + str(activities_count[act]) + ")"
                    decorations[trans] = {"label": label, "color": color}
        elif "performance" in variant:
            for act in decorations_single_contrib_trans:
                if act in trans_map:
                    trans = trans_map[act]
                    trans_values = list(
                        decorations_single_contrib_trans[act].values()
                    )
                    decorations[trans] = {"performance": mean(trans_values)}

    return decorations



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.conformance.alignments.dfg.variants import classic
from enum import Enum
from pm4py.util import exec_utils
from pm4py.objects.log.obj import EventLog, Trace
from typing import Optional, Dict, Any, Union, Tuple
from pm4py.util import typing
import pandas as pd


class Variants(Enum):
    CLASSIC = classic


def apply(
    obj: Union[EventLog, pd.DataFrame, Trace],
    dfg: Dict[Tuple[str, str], int],
    sa: Dict[str, int],
    ea: Dict[str, int],
    variant=Variants.CLASSIC,
    parameters: Optional[Dict[Any, Any]] = None,
) -> Union[typing.AlignmentResult, typing.ListAlignments]:
    """
    Applies the alignment algorithm provided a log/trace object, and a *connected* DFG

    Parameters
    --------------
    obj
        Event log / Trace
    dfg
        *Connected* directly-Follows Graph
    sa
        Start activities
    ea
        End activities
    variant
        Variant of the DFG alignments to be used. Possible values:
        - Variants.CLASSIC
    parameters
        Variant-specific parameters.

    Returns
    --------------
    ali
        Result of the alignment
    """
    return exec_utils.get_variant(variant).apply(
        obj, dfg, sa, ea, parameters=parameters
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import gzip
import os
import shutil
import tempfile


# this is ugly, should be done internally in the exporter...
def compress(file):
    """
    Compress a file in-place adding .gz suffix

    Parameters
    -----------
    file
        Uncompressed file

    Returns
    -----------
    compressed_file
        Compressed file path
    """
    extension = file.split(".")[-1] + ".gz"
    fp = tempfile.NamedTemporaryFile(suffix=extension)
    fp.close()
    with open(file, "rb") as f_in:
        with gzip.open(fp.name, "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)
    shutil.move(fp.name, file + ".gz")
    os.remove(file)
    return file



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.statistics.process_cube.variants import classic
from enum import Enum
from pm4py.util import exec_utils
import pandas as pd
from typing import Optional, Dict, Any, Tuple


class Variants(Enum):
    CLASSIC = classic


def apply(feature_table: pd.DataFrame, x_col: str, y_col: str, agg_col: str, variant=Variants.CLASSIC, parameters: Optional[Dict[Any, Any]] = None) -> Tuple[pd.DataFrame, Dict[Any, Any]]:
    """
    Constructs a process cube by slicing data along two dimensions (x_col, y_col) and aggregating a third (agg_col).
    Additionally:

    1) If x_col (or y_col) is an actual column in df, we do numeric binning.
       Otherwise, we do 'prefix-based' binning (include any column starting with x_col,
       and assign a row to that bin if >= 1).
    2) We return both the pivoted DataFrame and a dict associating each cell
       (x_bin, y_bin) -> set of case IDs.

    Parameters
    ----------
    feature_table : pd.DataFrame
        A feature table that must contain 'case:concept:name' and agg_col, plus
        the columns for x_col, y_col (if in numeric mode) or the columns that start
        with x_col, y_col (if in prefix mode).
    x_col : str
        The X dimension. If x_col in df.columns, use numeric binning. Otherwise, treat
        it as a prefix for 'prefix-based' binning.
    y_col : str
        The Y dimension. If y_col in df.columns, use numeric binning. Otherwise, treat
        it as a prefix for 'prefix-based' binning.
    agg_col : str
        The column to aggregate (mean, sum, etc.).
    variant
        Variant of the algorithm to be used, possible values:
        - Variants.CLASSIC
    parameters: Dict[Any, Any]
        Variant-specific parameters, including potentially:
        * Parameters.MAX_DIVISIONS_X: If x_col is numeric, how many bins to divide it into.
        * Parameters.MAX_DIVISIONS_Y: If y_col is numeric, how many bins to divide it into.
        * Parameters.AGGREGATION_FUNCTION: The aggregation function, e.g., 'mean', 'sum', 'min', 'max'.

    Returns
    -------
    pivot_df : pd.DataFrame
        A pivoted DataFrame representing the process cube, with x bins as rows
        and y bins as columns, containing aggregated values of agg_col.
    cell_case_dict : dict
        A dictionary mapping (x_bin, y_bin) -> set of case IDs that fall in that cell.
    """
    return exec_utils.get_variant(variant).apply(feature_table, x_col, y_col, agg_col, parameters=parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.conformance.tokenreplay.variants import token_replay, backwards
from enum import Enum
from pm4py.util import exec_utils
from typing import Optional, Dict, Any, Union
from pm4py.objects.log.obj import EventLog, EventStream
import pandas as pd
from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.util import typing


class Variants(Enum):
    TOKEN_REPLAY = token_replay
    BACKWARDS = backwards


VERSIONS = {Variants.TOKEN_REPLAY, Variants.BACKWARDS}
DEFAULT_VARIANT = Variants.TOKEN_REPLAY


def apply(
    log: Union[EventLog, EventStream, pd.DataFrame],
    net: PetriNet,
    initial_marking: Marking,
    final_marking: Marking,
    parameters: Optional[Dict[Any, Any]] = None,
    variant=DEFAULT_VARIANT,
) -> typing.ListAlignments:
    """
    Method to apply token-based replay

    Parameters
    -----------
    log
        Log
    net
        Petri net
    initial_marking
        Initial marking
    final_marking
        Final marking
    parameters
        Parameters of the algorithm, including:
            Parameters.ACTIVITY_KEY -> Activity key
    variant
        Variant of the algorithm to use:
            - Variants.TOKEN_REPLAY
            - Variants.BACKWARDS
    """
    if parameters is None:
        parameters = {}
    return exec_utils.get_variant(variant).apply(
        log, net, initial_marking, final_marking, parameters=parameters
    )


def get_diagnostics_dataframe(
    log: Union[EventLog, EventStream, pd.DataFrame],
    tbr_output: typing.ListAlignments,
    variant=DEFAULT_VARIANT,
    parameters: Optional[Dict[Any, Any]] = None,
) -> pd.DataFrame:
    """
    Gets the results of token-based replay in a dataframe

    Parameters
    --------------
    log
        Event log
    tbr_output
        Output of the token-based replay technique
    variant
        Variant of the algorithm to use:
            - Variants.TOKEN_REPLAY
            - Variants.BACKWARDS

    Returns
    --------------
    dataframe
        Diagnostics dataframe
    """
    if parameters is None:
        parameters = {}

    return exec_utils.get_variant(variant).get_diagnostics_dataframe(
        log, tbr_output, parameters=parameters
    )



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from typing import Optional, Dict, Any
from pm4py.objects.bpmn.obj import BPMN
from pm4py.util import exec_utils, constants, vis_utils
from enum import Enum
from copy import copy
import shutil
import tempfile
import importlib.resources


class Parameters(Enum):
    ENCODING = "encoding"
    IFRAME_WIDTH = "iframe_width"
    IFRAME_HEIGHT = "iframe_height"
    LOCAL_JUPYTER_FILE_NAME = "local_jupyter_file_name"


def get_html_file_contents():
    with importlib.resources.path(
        "pm4py.visualization.bpmn.util", "bpmnio_auto_layout.html"
    ) as p:
        with open(str(p), "r") as html_file:
            return html_file.read()


def apply(
    bpmn_graph: BPMN, parameters: Optional[Dict[Any, Any]] = None
) -> str:
    """
    Visualizes a BPMN graph by rendering it inside a HTML/Javascript file

    Parameters
    --------------
    bpmn_graph
        BPMN graph
    parameters
        Parameters of the algorithm, including:
        - Parameters.ENCODING => the encoding of the HTML to be used

    Returns
    ---------------
    tmp_file_path
        Path to the HTML file
    """
    if parameters is None:
        parameters = {}

    encoding = exec_utils.get_param_value(
        Parameters.ENCODING, parameters, constants.DEFAULT_ENCODING
    )

    from pm4py.objects.bpmn.exporter.variants import etree as bpmn_xml_exporter

    export_parameters = copy(parameters)
    xml_stri = bpmn_xml_exporter.get_xml_string(
        bpmn_graph, parameters=export_parameters
    )
    xml_stri = (
        xml_stri.decode(encoding)
        .replace("\t", "")
        .replace("\r", "")
        .replace("\n", "")
    )

    F = tempfile.NamedTemporaryFile(suffix=".html")
    F.close()

    F = open(F.name, "w", encoding=encoding)
    F.write(get_html_file_contents().replace("REPLACE", xml_stri))
    F.close()

    return F.name


def view(temp_file_name, parameters=None):
    """
    View the SNA visualization on the screen

    Parameters
    -------------
    temp_file_name
        Temporary file name
    parameters
        Possible parameters of the algorithm
    """
    if parameters is None:
        parameters = {}

    if constants.DEFAULT_ENABLE_VISUALIZATIONS_VIEW:
        iframe_width = exec_utils.get_param_value(
            Parameters.IFRAME_WIDTH, parameters, 900
        )
        iframe_height = exec_utils.get_param_value(
            Parameters.IFRAME_HEIGHT, parameters, 600
        )
        local_jupyter_file_name = exec_utils.get_param_value(
            Parameters.LOCAL_JUPYTER_FILE_NAME,
            parameters,
            "jupyter_bpmn_vis.html",
        )

        if vis_utils.check_visualization_inside_jupyter():
            from IPython.display import IFrame

            shutil.copyfile(temp_file_name, local_jupyter_file_name)
            iframe = IFrame(
                local_jupyter_file_name,
                width=iframe_width,
                height=iframe_height,
            )
            from IPython.display import display

            return display(iframe)
        else:
            vis_utils.open_opsystem_image_viewer(temp_file_name)


def save(temp_file_name, dest_file, parameters=None):
    """
    Save the SNA visualization from a temporary file to a well-defined destination file

    Parameters
    -------------
    temp_file_name
        Temporary file name
    dest_file
        Destination file
    parameters
        Possible parameters of the algorithm
    """
    if parameters is None:
        parameters = {}

    shutil.copyfile(temp_file_name, dest_file)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

import pandas as pd
from pm4py.objects.log.obj import EventLog, EventStream, Trace
from typing import Union, Optional, Dict, Tuple, List, Any
from pm4py.utils import get_properties, constants, check_is_pandas_dataframe
from pm4py.utils import __event_log_deprecation_warning
from pm4py.objects.ocel.obj import OCEL
from tempfile import NamedTemporaryFile
from copy import copy
from pm4py.objects.petri_net.obj import PetriNet, Marking


def openai_query(
    prompt: str,
    api_key: Optional[str] = None,
    openai_model: Optional[str] = None,
    api_url: Optional[str] = None,
    extra_payload: Optional[Dict[Any, Any]] = None,
    **kwargs
) -> str:
    """
    Executes the provided prompt, obtaining the answer from the OpenAI APIs.

    :param prompt: The prompt to be executed.
    :param api_key: (Optional) OpenAI API key.
    :param openai_model: (Optional) OpenAI model to be used (default: "gpt-3.5-turbo").
    :param api_url: (Optional) OpenAI API URL.
    :param extra_payload: (Optional) Additional payload fields merged into the request body.
    :param **kwargs: Additional parameters to pass to the OpenAI API.
    :return: The response from the OpenAI API as a string.

    .. code-block:: python3

        import pm4py

        resp = pm4py.llm.openai_query('What is the result of 3+3?', api_key="sk-382393", openai_model="gpt-3.5-turbo")
        print(resp)
    """
    parameters = copy(kwargs) if kwargs is not None else {}
    if api_url is not None:
        parameters["api_url"] = api_url
    if api_key is not None:
        parameters["api_key"] = api_key
    if openai_model is not None:
        parameters["openai_model"] = openai_model
    if extra_payload is not None:
        parameters["extra_payload"] = extra_payload

    from pm4py.algo.querying.llm.connectors import openai as perform_query

    return perform_query.apply(prompt, parameters=parameters)


def google_query(
    prompt: str,
    api_key: Optional[str] = None,
    model: Optional[str] = None,
    extra_payload: Optional[Dict[Any, Any]] = None,
    **kwargs
) -> str:
    """
    Executes the provided prompt, obtaining the answer from the Google APIs.

    :param prompt: prompt that should be executed
    :param api_key: API key
    :param model: Model to be used (default: gemini-1.5-flash-002)
    :param extra_payload: (Optional) Additional payload fields merged into the request body.
    :rtype: ``str``

    .. code-block:: python3

        import pm4py

        resp = pm4py.llm.google_query('what is the result of 3+3?', api_key="sk-382393", model="gemini-1.5-flash-002")
        print(resp)
    """
    parameters = copy(kwargs) if kwargs is not None else {}
    if api_key is not None:
        parameters["api_key"] = api_key
    if model is not None:
        parameters["google_model"] = model
    if extra_payload is not None:
        parameters["extra_payload"] = extra_payload

    from pm4py.algo.querying.llm.connectors import google as perform_query

    return perform_query.apply(prompt, parameters=parameters)


def anthropic_query(
    prompt: str,
    api_key: Optional[str] = None,
    model: Optional[str] = None,
    extra_payload: Optional[Dict[Any, Any]] = None,
    **kwargs
) -> str:
    """
    Executes the provided prompt, obtaining the answer from the Google APIs.

    :param prompt: prompt that should be executed
    :param api_key: API key
    :param model: Model to be used (default: claude-3-5-sonnet-20241022)
    :param extra_payload: (Optional) Additional payload fields merged into the request body.
    :rtype: ``str``

    .. code-block:: python3

        import pm4py

        resp = pm4py.llm.anthropic_query('what is the result of 3+3?', api_key="sk-382393", model="claude-3-5-sonnet-20241022")
        print(resp)
    """
    parameters = copy(kwargs) if kwargs is not None else {}
    if api_key is not None:
        parameters["api_key"] = api_key
    if model is not None:
        parameters["anthropic_model"] = model
    if extra_payload is not None:
        parameters["extra_payload"] = extra_payload

    from pm4py.algo.querying.llm.connectors import anthropic as perform_query

    return perform_query.apply(prompt, parameters=parameters)


def abstract_dfg(
    log_obj: Union[pd.DataFrame, EventLog, EventStream],
    max_len: int = constants.OPENAI_MAX_LEN,
    include_performance: bool = True,
    relative_frequency: bool = False,
    response_header: bool = True,
    primary_performance_aggregation: str = "mean",
    secondary_performance_aggregation: Optional[str] = None,
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name",
) -> str:
    """
    Obtains the DFG (Directly-Follows Graph) abstraction of a traditional event log.

    :param log_obj: The log object to abstract.
    :param max_len: Maximum length of the string abstraction (default: constants.OPENAI_MAX_LEN).
    :param include_performance: Whether to include the performance of the paths in the abstraction.
    :param relative_frequency: Whether to use relative instead of absolute frequency of the paths.
    :param response_header: Whether to include a short header before the paths, describing the abstraction.
    :param primary_performance_aggregation: Primary aggregation method for the arc's performance (default: "mean"). Other options: "median", "min", "max", "sum", "stdev".
    :param secondary_performance_aggregation: (Optional) Secondary aggregation method for the arc's performance (default: None). Other options: "mean", "median", "min", "max", "sum", "stdev".
    :param activity_key: The column name to be used as activity.
    :param timestamp_key: The column name to be used as timestamp.
    :param case_id_key: The column name to be used as case identifier.
    :return: The DFG abstraction as a string.

    .. code-block:: python3

        import pm4py

        log = pm4py.read_xes("tests/input_data/roadtraffic100traces.xes")
        print(pm4py.llm.abstract_dfg(log))
    """
    __event_log_deprecation_warning(log_obj)

    parameters = get_properties(
        log_obj,
        activity_key=activity_key,
        timestamp_key=timestamp_key,
        case_id_key=case_id_key,
    )
    parameters["max_len"] = max_len
    parameters["include_performance"] = include_performance
    parameters["relative_frequency"] = relative_frequency
    parameters["response_header"] = response_header
    parameters["primary_performance_aggregation"] = (
        primary_performance_aggregation
    )
    parameters["secondary_performance_aggregation"] = (
        secondary_performance_aggregation
    )

    from pm4py.algo.querying.llm.abstractions import log_to_dfg_descr

    return log_to_dfg_descr.apply(log_obj, parameters=parameters)


def abstract_variants(
    log_obj: Union[pd.DataFrame, EventLog, EventStream],
    max_len: int = constants.OPENAI_MAX_LEN,
    include_performance: bool = True,
    relative_frequency: bool = False,
    response_header: bool = True,
    primary_performance_aggregation: str = "mean",
    secondary_performance_aggregation: Optional[str] = None,
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name",
) -> str:
    """
    Obtains the variants abstraction of a traditional event log.

    :param log_obj: The log object to abstract.
    :param max_len: Maximum length of the string abstraction (default: constants.OPENAI_MAX_LEN).
    :param include_performance: Whether to include the performance of the variants in the abstraction.
    :param relative_frequency: Whether to use relative instead of absolute frequency of the variants.
    :param response_header: Whether to include a short header before the variants, describing the abstraction.
    :param primary_performance_aggregation: Primary aggregation method for the variants' performance (default: "mean"). Other options: "median", "min", "max", "sum", "stdev".
    :param secondary_performance_aggregation: (Optional) Secondary aggregation method for the variants' performance (default: None). Other options: "mean", "median", "min", "max", "sum", "stdev".
    :param activity_key: The column name to be used as activity.
    :param timestamp_key: The column name to be used as timestamp.
    :param case_id_key: The column name to be used as case identifier.
    :return: The variants abstraction as a string.

    .. code-block:: python3

        import pm4py

        log = pm4py.read_xes("tests/input_data/roadtraffic100traces.xes")
        print(pm4py.llm.abstract_variants(log))
    """
    __event_log_deprecation_warning(log_obj)

    parameters = get_properties(
        log_obj,
        activity_key=activity_key,
        timestamp_key=timestamp_key,
        case_id_key=case_id_key,
    )
    parameters["max_len"] = max_len
    parameters["include_performance"] = include_performance
    parameters["relative_frequency"] = relative_frequency
    parameters["response_header"] = response_header
    parameters["primary_performance_aggregation"] = (
        primary_performance_aggregation
    )
    parameters["secondary_performance_aggregation"] = (
        secondary_performance_aggregation
    )

    from pm4py.algo.querying.llm.abstractions import log_to_variants_descr

    return log_to_variants_descr.apply(log_obj, parameters=parameters)


def abstract_ocel(ocel: OCEL, include_timestamps: bool = True) -> str:
    """
    Obtains the abstraction of an object-centric event log, including the list of events and the objects of the OCEL.

    :param ocel: The object-centric event log to abstract.
    :param include_timestamps: Whether to include timestamp information in the abstraction.
    :return: The OCEL abstraction as a string.

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel("tests/input_data/ocel/example_log.jsonocel")
        print(pm4py.llm.abstract_ocel(ocel))
    """
    parameters = {}
    parameters["include_timestamps"] = include_timestamps

    from pm4py.algo.transformation.ocel.description import (
        algorithm as ocel_description,
    )

    return ocel_description.apply(ocel, parameters=parameters)


def abstract_ocel_ocdfg(
    ocel: OCEL,
    include_header: bool = True,
    include_timestamps: bool = True,
    max_len: int = constants.OPENAI_MAX_LEN,
) -> str:
    """
    Obtains the abstraction of an object-centric event log, representing the object-centric directly-follows graph in text.

    :param ocel: The object-centric event log to abstract.
    :param include_header: Whether to include a header in the abstraction.
    :param include_timestamps: Whether to include timestamp information in the abstraction.
    :param max_len: Maximum length of the abstraction (default: constants.OPENAI_MAX_LEN).
    :return: The object-centric DFG abstraction as a string.

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel("tests/input_data/ocel/example_log.jsonocel")
        print(pm4py.llm.abstract_ocel_ocdfg(ocel))
    """
    parameters = {}
    parameters["include_header"] = include_header
    parameters["include_timestamps"] = include_timestamps
    parameters["max_len"] = max_len

    from pm4py.algo.querying.llm.abstractions import ocel_ocdfg_descr

    return ocel_ocdfg_descr.apply(ocel, parameters=parameters)


def abstract_ocel_features(
    ocel: OCEL,
    obj_type: str,
    include_header: bool = True,
    max_len: int = constants.OPENAI_MAX_LEN,
    debug: bool = False,
    enable_object_lifecycle_paths: bool = True,
) -> str:
    """
    Obtains the abstraction of an object-centric event log, representing the features and their values in text.

    :param ocel: The object-centric event log to abstract.
    :param obj_type: The object type to consider in feature extraction.
    :param include_header: Whether to include a header in the abstraction.
    :param max_len: Maximum length of the abstraction (default: constants.OPENAI_MAX_LEN).
    :param debug: Enables debugging mode, providing insights into feature extraction steps.
    :param enable_object_lifecycle_paths: Enables the "lifecycle paths" feature in the abstraction.
    :return: The OCEL features abstraction as a string.

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel("tests/input_data/ocel/example_log.jsonocel")
        print(pm4py.llm.abstract_ocel_features(ocel, obj_type="Resource"))
    """
    parameters = {}
    parameters["include_header"] = include_header
    parameters["max_len"] = max_len
    parameters["debug"] = debug
    parameters["enable_object_lifecycle_paths"] = enable_object_lifecycle_paths

    from pm4py.algo.querying.llm.abstractions import ocel_fea_descr

    return ocel_fea_descr.apply(ocel, obj_type, parameters=parameters)


def abstract_event_stream(
    log_obj: Union[pd.DataFrame, EventLog, EventStream],
    max_len: int = constants.OPENAI_MAX_LEN,
    response_header: bool = True,
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name",
) -> str:
    """
    Obtains the event stream abstraction of a traditional event log.

    :param log_obj: The log object to abstract.
    :param max_len: Maximum length of the string abstraction (default: constants.OPENAI_MAX_LEN).
    :param response_header: Whether to include a short header before the event stream, describing the abstraction.
    :param activity_key: The column name to be used as activity.
    :param timestamp_key: The column name to be used as timestamp.
    :param case_id_key: The column name to be used as case identifier.
    :return: The event stream abstraction as a string.

    .. code-block:: python3

        import pm4py

        log = pm4py.read_xes("tests/input_data/roadtraffic100traces.xes")
        print(pm4py.llm.abstract_event_stream(log))
    """
    __event_log_deprecation_warning(log_obj)

    parameters = get_properties(
        log_obj,
        activity_key=activity_key,
        timestamp_key=timestamp_key,
        case_id_key=case_id_key,
    )
    parameters["max_len"] = max_len
    parameters["response_header"] = response_header

    from pm4py.algo.querying.llm.abstractions import stream_to_descr

    return stream_to_descr.apply(log_obj, parameters=parameters)


def abstract_petri_net(
    net: PetriNet, im: Marking, fm: Marking, response_header: bool = True
) -> str:
    """
    Obtains an abstraction of a Petri net.

    :param net: The Petri net to abstract.
    :param im: The initial marking of the Petri net.
    :param fm: The final marking of the Petri net.
    :param response_header: Whether to include a header in the abstraction.
    :return: The Petri net abstraction as a string.

    .. code-block:: python3

        import pm4py

        net, im, fm = pm4py.read_pnml('tests/input_data/running-example.pnml')
        print(pm4py.llm.abstract_petri_net(net, im, fm))
    """
    parameters = {}
    parameters["response_header"] = response_header

    from pm4py.algo.querying.llm.abstractions import net_to_descr

    return net_to_descr.apply(net, im, fm, parameters=parameters)


def abstract_log_attributes(
    log_obj: Union[pd.DataFrame, EventLog, EventStream],
    max_len: int = constants.OPENAI_MAX_LEN,
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name",
) -> str:
    """
    Abstracts the attributes of a log by reporting their names, types, and top values.

    :param log_obj: The log object whose attributes are to be abstracted.
    :param max_len: Maximum length of the string abstraction (default: constants.OPENAI_MAX_LEN).
    :param activity_key: The column name to be used as activity.
    :param timestamp_key: The column name to be used as timestamp.
    :param case_id_key: The column name to be used as case identifier.
    :return: The log attributes abstraction as a string.

    .. code-block:: python3

        import pm4py

        log = pm4py.read_xes("tests/input_data/roadtraffic100traces.xes")
        print(pm4py.llm.abstract_log_attributes(log))
    """
    __event_log_deprecation_warning(log_obj)

    parameters = get_properties(
        log_obj,
        activity_key=activity_key,
        timestamp_key=timestamp_key,
        case_id_key=case_id_key,
    )
    parameters["max_len"] = max_len

    from pm4py.algo.querying.llm.abstractions import log_to_cols_descr

    return log_to_cols_descr.apply(log_obj, parameters=parameters)


def abstract_log_features(
    log_obj: Union[pd.DataFrame, EventLog, EventStream],
    max_len: int = constants.OPENAI_MAX_LEN,
    include_header: bool = True,
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name",
) -> str:
    """
    Abstracts the machine learning features obtained from a log by reporting the top features until the desired length is achieved.

    :param log_obj: The log object from which to extract features.
    :param max_len: Maximum length of the string abstraction (default: constants.OPENAI_MAX_LEN).
    :param include_header: Whether to include a header in the abstraction.
    :param activity_key: The column name to be used as activity.
    :param timestamp_key: The column name to be used as timestamp.
    :param case_id_key: The column name to be used as case identifier.
    :return: The log features abstraction as a string.

    .. code-block:: python3

        import pm4py

        log = pm4py.read_xes("tests/input_data/roadtraffic100traces.xes")
        print(pm4py.llm.abstract_log_features(log))
    """
    __event_log_deprecation_warning(log_obj)

    parameters = get_properties(
        log_obj,
        activity_key=activity_key,
        timestamp_key=timestamp_key,
        case_id_key=case_id_key,
    )
    parameters["max_len"] = max_len
    parameters["include_header"] = include_header

    from pm4py.algo.querying.llm.abstractions import log_to_fea_descr

    return log_to_fea_descr.apply(log_obj, parameters=parameters)


def abstract_temporal_profile(
    temporal_profile: Dict[Tuple[str, str], Tuple[float, float]],
    include_header: bool = True,
) -> str:
    """
    Abstracts a temporal profile model into a descriptive string.

    :param temporal_profile: The temporal profile model to abstract.
    :param include_header: Whether to include a header in the abstraction describing the temporal profile.
    :return: The temporal profile abstraction as a string.

    .. code-block:: python3

        import pm4py

        log = pm4py.read_xes("tests/input_data/roadtraffic100traces.xes", return_legacy_log_object=True)
        temporal_profile = pm4py.discover_temporal_profile(log)
        text_abstr = pm4py.llm.abstract_temporal_profile(temporal_profile, include_header=True)
        print(text_abstr)
    """
    parameters = {}
    parameters["include_header"] = include_header

    from pm4py.algo.querying.llm.abstractions import tempprofile_to_descr

    return tempprofile_to_descr.apply(temporal_profile, parameters=parameters)


def abstract_case(
    case: Trace,
    include_case_attributes: bool = True,
    include_event_attributes: bool = True,
    include_timestamp: bool = True,
    include_header: bool = True,
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
) -> str:
    """
    Textually abstracts a single case from an event log.

    :param case: The case object to abstract.
    :param include_case_attributes: Whether to include attributes at the case level.
    :param include_event_attributes: Whether to include attributes at the event level.
    :param include_timestamp: Whether to include event timestamps in the abstraction.
    :param include_header: Whether to include a header in the abstraction.
    :param activity_key: The column name to be used as activity.
    :param timestamp_key: The column name to be used as timestamp.
    :return: The case abstraction as a string.

    .. code-block:: python3

        import pm4py

        log = pm4py.read_xes("tests/input_data/roadtraffic100traces.xes", return_legacy_log_object=True)
        print(pm4py.llm.abstract_case(log[0]))
    """
    parameters = {}
    parameters["include_case_attributes"] = include_case_attributes
    parameters["include_event_attributes"] = include_event_attributes
    parameters["include_timestamp"] = include_timestamp
    parameters["include_header"] = include_header
    parameters[constants.PARAMETER_CONSTANT_ACTIVITY_KEY] = activity_key
    parameters[constants.PARAMETER_CONSTANT_TIMESTAMP_KEY] = timestamp_key

    from pm4py.algo.querying.llm.abstractions import case_to_descr

    return case_to_descr.apply(case, parameters=parameters)


def abstract_declare(declare_model, include_header: bool = True) -> str:
    """
    Textually abstracts a DECLARE model.

    :param declare_model: The DECLARE model to abstract.
    :param include_header: Whether to include a header in the abstraction.
    :return: The DECLARE model abstraction as a string.

    .. code-block:: python3

        import pm4py

        log = pm4py.read_xes("tests/input_data/roadtraffic100traces.xes", return_legacy_log_object=True)
        log_ske = pm4py.discover_declare(log)
        print(pm4py.llm.abstract_declare(log_ske))
    """
    parameters = {}
    parameters["include_header"] = include_header

    from pm4py.algo.querying.llm.abstractions import declare_to_descr

    return declare_to_descr.apply(declare_model, parameters=parameters)


def abstract_log_skeleton(log_skeleton, include_header: bool = True) -> str:
    """
    Textually abstracts a log skeleton process model.

    :param log_skeleton: The log skeleton to abstract.
    :param include_header: Whether to include a header in the abstraction.
    :return: The log skeleton abstraction as a string.

    .. code-block:: python3

        import pm4py

        log = pm4py.read_xes("tests/input_data/roadtraffic100traces.xes", return_legacy_log_object=True)
        log_ske = pm4py.discover_log_skeleton(log)
        print(pm4py.llm.abstract_log_skeleton(log_ske))
    """
    parameters = {}
    parameters["include_header"] = include_header

    from pm4py.algo.querying.llm.abstractions import logske_to_descr

    return logske_to_descr.apply(log_skeleton, parameters=parameters)


def __execute_prompt_to_db_query(
    log: pd.DataFrame,
    prompt: str,
    executor=openai_query,
    execute_query: bool = True,
    **kwargs
) -> Union[str, pd.DataFrame]:
    """
    Internal method to retrieve (and execute) a SQL query corresponding to a prompt
    """
    import duckdb

    if check_is_pandas_dataframe(log):
        kwargs["table_name"] = "dataframe"
        dataframe = log

    response = executor(prompt, **kwargs)

    sql_query = ""
    if "```sql" in response:
        sql_query = response.split("```sql")[1].split("```")[0]
    sql_query = sql_query.strip()

    if not sql_query:
        raise Exception(
            "The response does not contain a valid SQL query: " + response
        )

    if execute_query:
        result = duckdb.query(sql_query).to_df()

        if check_is_pandas_dataframe(log):
            return result

    return sql_query


def nlp_to_log_query(
    log: pd.DataFrame,
    query: str,
    executor=openai_query,
    obtain_query: bool = True,
    execute_query: bool = True,
    **kwargs
) -> Union[str, pd.DataFrame]:
    """
    Translates a natural language statement into a database (SQL) query executable against the event log.

    :param log: event log object
    :param query: query expressed in natural language
    :param executor: the connector to the LLM (e.g., pm4py.llm.openai_query)
    :param obtain_query: executes the prompt, to transform the natural statements into a database (SQL) query
    :param execute_query: executes the database (SQL) query against the event data
    :param kwargs: additional keyword arguments to the method
    :rtype: ``Union[str, pd.DataFrame]``

    .. code-block:: python3

        import pm4py

        log = pm4py.read_xes("tests/input_data/running-example.xes")
        resp = pm4py.llm.nlp_to_log_query(log, "How many cases are contained in the event log?", api_key="sk-5HNn")
        print(resp)
    """
    from pm4py.algo.querying.llm.injection import (
        algorithm as domain_knowledge_injector,
    )

    prompt = (
        "Could you provide a database query for the following question?\n\n"
        + query
    )
    prompt += domain_knowledge_injector.apply(log, parameters=kwargs)
    prompt += "\n\nPlease include the SQL query between the ```sql and the ``` tags.\n\n"

    if obtain_query:
        return __execute_prompt_to_db_query(
            log,
            prompt,
            executor=executor,
            execute_query=execute_query,
            **kwargs
        )

    return prompt


def nlp_to_log_filter(
    log: pd.DataFrame,
    filter_query: str,
    executor=openai_query,
    obtain_query: bool = True,
    execute_query: bool = True,
    **kwargs
) -> Union[str, pd.DataFrame]:
    """
    Translates a filtering query expressed in natural language into a database (SQL) query that is used to filter the
    event log.

    :param log: event log object
    :param filter_query: filtering query expressed in natural language
    :param executor: the connector to the LLM (e.g., pm4py.llm.openai_query)
    :param obtain_query: executes the prompt, to transform the natural statements into a database (SQL) query
    :param execute_query: executes the database (SQL) query against the event data
    :param kwargs: additional keyword arguments to the method
    :rtype: ``Union[str, pd.DataFrame]``

    .. code-block:: python3

        import pm4py

        log = pm4py.read_xes("tests/input_data/running-example.xes")
        resp = pm4py.llm.nlp_to_log_filter(log, "There is an event with activity: pay compensation", api_key="sk-5HNn")
        print(resp)
    """
    from pm4py.algo.querying.llm.injection import (
        algorithm as domain_knowledge_injector,
    )

    prompt = (
        "Could you provide a database query to filter all the events of the cases for which at least an event (row) is satisfying the following filtering query?\n\n" +
        filter_query)
    prompt += domain_knowledge_injector.apply(log, parameters=kwargs)
    prompt += "\n\nPlease include the SQL query between the ```sql and the ``` tags.\n\n"

    if obtain_query:
        return __execute_prompt_to_db_query(
            log,
            prompt,
            executor=executor,
            execute_query=execute_query,
            **kwargs
        )

    return prompt


def automated_hypotheses_formulation(
    dataframe: pd.DataFrame,
    executor=openai_query,
    obtain_query: bool = True,
    execute_query: bool = True,
    max_len=10000,
    **kwargs
) -> Union[str, List[Tuple[str, str, Union[str, None]]]]:
    """
    Automatically formulate some hypotheses on the event data.
    The result of this method is either:
    - The prompt (to be executed manually against the LLM) when obtain_query=False
    - A list of different hypotheses. Each hypothesis comes with a description (position 0), the SQL query (position 1),
    and (if execute_query=True) the result of the execution of the query in position 2.

    :param dataframe: event log object
    :param executor: the connector to the LLM (e.g., pm4py.llm.openai_query)
    :param obtain_query: executes the prompt and get the hypotheses along with the corresponding SQL queries
    :param execute_query: executes the obtained SQL queries and stores the results
    :param kwargs: additional keyword arguments to the method
    :param max_len: maximum length of the obtained prompt
    :rtype: ``Union[str, List[Tuple[str, str, Union[str, None]]]]``

    .. code-block:: python3

        import pm4py

        log = pm4py.read_xes("tests/input_data/running-example.xes")
        result = pm4py.llm.automated_hypotheses_formulation(log, api_key="sk-5HN")
        print(result)
    """
    import duckdb
    from pm4py.algo.querying.llm.abstractions import log_to_cols_descr

    from pm4py.algo.querying.llm.injection import (
        algorithm as domain_knowledge_injector,
    )

    prompt = "Could you formulate some hypotheses over the event data? The directly-follows graph, the list of attributes, and additional process mining/database knowledge follow.\n\n"
    prompt += abstract_dfg(dataframe, max_len=int(max_len / 3))
    prompt += "\n\n"
    prompt += log_to_cols_descr.apply(
        dataframe, parameters={"max_len": int(max_len / 3)}
    )
    prompt += "\n\n"
    prompt += domain_knowledge_injector.apply(dataframe, parameters=kwargs)
    prompt += "Please provide each hypothesis along with a SQL query. Please include the SQL queries between the ```sql and the ``` tags.\n"
    prompt += "Can you please include also a description of each hypothesis between some <description> and </description> tags?\n\n"

    if obtain_query:
        response = executor(prompt, **kwargs)

        obtained_queries = response.split("```sql")
        obtained_descriptions = response.split("<description>")

        if (
            len(obtained_queries) > 1
            and len(obtained_descriptions) > 1
            and len(obtained_queries) == len(obtained_descriptions)
        ):
            del obtained_queries[0]
            del obtained_descriptions[0]

            obtained_queries = [x.split("```")[0] for x in obtained_queries]
            obtained_descriptions = [
                x.split("</description>")[0].strip()
                for x in obtained_descriptions
            ]
            obtained_results = [None] * len(obtained_descriptions)

            if execute_query:
                for i in range(len(obtained_queries)):
                    try:
                        obtained_results[i] = duckdb.sql(
                            obtained_queries[i]
                        ).to_df()
                    except BaseException:
                        print("Exception executing query nr. %d" % i)

            return [
                (
                    obtained_descriptions[i],
                    obtained_queries[i],
                    obtained_results[i],
                )
                for i in range(len(obtained_queries))
            ]

    return prompt


def explain_visualization(
    vis_saver, *args, connector=openai_query, **kwargs
) -> str:
    """
    Explains a process mining visualization using LLMs by saving it as a .png image and providing the image to the Large Language Model along with a description.

    :param vis_saver: The visualizer function used to save the visualization to disk.
    :param args: Positional arguments required by the visualizer function.
    :param connector: (Optional) The connector method to communicate with the large language model (default: openai_query).
    :param **kwargs: Additional keyword arguments for the visualizer function or the connector (e.g., annotations, API key).
    :return: The explanation of the visualization as a string.

    .. code-block:: python3

        import pm4py

        log = pm4py.read_xes("tests/input_data/running-example.xes")
        descr = pm4py.llm.explain_visualization(pm4py.save_vis_dotted_chart, log, api_key="sk-5HN", show_legend=False)
        print(descr)
    """
    F = NamedTemporaryFile(suffix=".png")
    image_path = F.name
    F.close()

    description = vis_saver(*args, image_path, **kwargs)

    parameters = copy(kwargs) if kwargs is not None else {}
    parameters["image_path"] = image_path

    return connector(
        "Could you explain the included process mining visualization?\n\n"
        + description,
        **parameters
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from copy import copy
from enum import Enum
from typing import Optional, Dict, Any, Tuple

import pandas as pd

from pm4py.algo.discovery.dfg import algorithm as dfg_alg
from pm4py.objects.conversion.heuristics_net import converter as hn_conv_alg
from pm4py.objects.conversion.log import converter as log_converter
from pm4py.objects.heuristics_net import defaults
from pm4py.objects.heuristics_net.obj import HeuristicsNet
from pm4py.objects.heuristics_net.node import Node
from pm4py.objects.log.obj import EventLog
from pm4py.objects.log.util import interval_lifecycle
from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.statistics.attributes.log import get as log_attributes
from pm4py.statistics.concurrent_activities.log import get as conc_act_get
from pm4py.statistics.end_activities.log import get as log_ea
from pm4py.statistics.eventually_follows.log import get as efg_get
from pm4py.statistics.service_time.log import get as soj_get
from pm4py.statistics.start_activities.log import get as log_sa
from pm4py.util import exec_utils, constants, xes_constants as xes


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    DEPENDENCY_THRESH = "dependency_thresh"
    AND_MEASURE_THRESH = "and_measure_thresh"
    MIN_ACT_COUNT = "min_act_count"
    MIN_DFG_OCCURRENCES = "min_dfg_occurrences"
    HEU_NET_DECORATION = "heu_net_decoration"

def is_polars_lazyframe(df: Any) -> bool:
    """Return True if the provided dataframe is a Polars LazyFrame."""
    df_type = str(type(df)).lower()
    return "polars" in df_type and "lazyframe" in df_type


def apply(
    log: EventLog, parameters: Optional[Dict[Any, Any]] = None
) -> Tuple[PetriNet, Marking, Marking]:
    """
    Discovers a Petri net using the Heuristics Miner ++ algorithm

    Implements the approach described in
    Burattin, Andrea, and Alessandro Sperduti. "Heuristics Miner for Time Intervals." ESANN. 2010.

    https://andrea.burattin.net/public-files/publications/2010-esann-slides.pdf

    Parameters
    --------------
    log
        Event log
    parameters
        Parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY
        - Parameters.START_TIMESTAMP_KEY
        - Parameters.TIMESTAMP_KEY
        - Parameters.DEPENDENCY_THRESH
        - Parameters.AND_MEASURE_THRESH
        - Parameters.MIN_ACT_COUNT
        - Parameters.MIN_DFG_OCCURRENCES
        - Parameters.HEU_NET_DECORATION

    Returns
    --------------
    net
        Petri net
    im
        Initial marking
    fm
        Final marking
    """
    heu_net = apply_heu(log, parameters=parameters)
    net, im, fm = hn_conv_alg.apply(heu_net, parameters=parameters)
    return net, im, fm


def apply_pandas(
    df: pd.DataFrame, parameters: Optional[Dict[Any, Any]] = None
) -> Tuple[PetriNet, Marking, Marking]:
    """
    Discovers a Petri net using the Heuristics Miner ++ algorithm

    Implements the approach described in
    Burattin, Andrea, and Alessandro Sperduti. "Heuristics Miner for Time Intervals." ESANN. 2010.

    https://andrea.burattin.net/public-files/publications/2010-esann-slides.pdf

    Parameters
    --------------
    df
        Dataframe
    parameters
        Parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY
        - Parameters.START_TIMESTAMP_KEY
        - Parameters.TIMESTAMP_KEY
        - Parameters.CASE_ID_KEY
        - Parameters.DEPENDENCY_THRESH
        - Parameters.AND_MEASURE_THRESH
        - Parameters.MIN_ACT_COUNT
        - Parameters.MIN_DFG_OCCURRENCES
        - Parameters.HEU_NET_DECORATION

    Returns
    --------------
    net
        Petri net
    im
        Initial marking
    fm
        Final marking
    """
    heu_net = apply_heu_pandas(df, parameters=parameters)
    net, im, fm = hn_conv_alg.apply(heu_net, parameters=parameters)
    return net, im, fm


def apply_heu(
    log: EventLog, parameters: Optional[Dict[Any, Any]] = None
) -> HeuristicsNet:
    """
    Discovers an heuristics net using the Heuristics Miner ++ algorithm

    Implements the approach described in
    Burattin, Andrea, and Alessandro Sperduti. "Heuristics Miner for Time Intervals." ESANN. 2010.

    https://andrea.burattin.net/public-files/publications/2010-esann-slides.pdf

    Parameters
    --------------
    log
        Event log
    parameters
        Parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY
        - Parameters.START_TIMESTAMP_KEY
        - Parameters.TIMESTAMP_KEY
        - Parameters.DEPENDENCY_THRESH
        - Parameters.AND_MEASURE_THRESH
        - Parameters.MIN_ACT_COUNT
        - Parameters.MIN_DFG_OCCURRENCES
        - Parameters.HEU_NET_DECORATION

    Returns
    --------------
    heu_net
        Heuristics net
    """
    if parameters is None:
        parameters = {}

    log = log_converter.apply(
        log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters
    )
    log = interval_lifecycle.to_interval(log, parameters=parameters)
    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY, parameters, None
    )
    if start_timestamp_key is None:
        start_timestamp_key = xes.DEFAULT_START_TIMESTAMP_KEY
        parameters = copy(parameters)
        parameters[Parameters.START_TIMESTAMP_KEY] = start_timestamp_key
    (
        start_activities,
        end_activities,
        activities_occurrences,
        dfg,
        performance_dfg,
        sojourn_time,
        concurrent_activities,
    ) = discover_abstraction_log(log, parameters=parameters)
    return discover_heu_net_plus_plus(
        start_activities,
        end_activities,
        activities_occurrences,
        dfg,
        performance_dfg,
        sojourn_time,
        concurrent_activities,
        parameters=parameters,
    )


def discover_abstraction_log(
    log: EventLog, parameters: Optional[Dict[Any, Any]] = None
) -> Tuple[Any, Any, Any, Any, Any, Any, Any]:
    """
    Discovers an abstraction from a log that is useful for the Heuristics Miner ++ algorithm

    Parameters
    --------------
    log
        Event log
    parameters
        Parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY
        - Parameters.START_TIMESTAMP_KEY
        - Parameters.TIMESTAMP_KEY
        - Parameters.CASE_ID_KEY

    Returns
    --------------
    start_activities
        Start activities
    end_activities
        End activities
    activities_occurrences
        Activities along with their number of occurrences
    dfg
        Directly-follows graph
    performance_dfg
        (Performance) Directly-follows graph
    sojourn_time
        Sojourn time for each activity
    concurrent_activities
        Concurrent activities
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes.DEFAULT_NAME_KEY
    )
    start_activities = log_sa.get_start_activities(log, parameters=parameters)
    end_activities = log_ea.get_end_activities(log, parameters=parameters)
    activities_occurrences = log_attributes.get_attribute_values(
        log, activity_key, parameters=parameters
    )
    efg_parameters = copy(parameters)
    efg_parameters[efg_get.Parameters.KEEP_FIRST_FOLLOWING] = True
    dfg = efg_get.apply(log, parameters=efg_parameters)
    performance_dfg = dfg_alg.apply(
        log, variant=dfg_alg.Variants.PERFORMANCE, parameters=parameters
    )
    sojourn_time = soj_get.apply(log, parameters=parameters)
    concurrent_activities = conc_act_get.apply(log, parameters=parameters)
    return (
        start_activities,
        end_activities,
        activities_occurrences,
        dfg,
        performance_dfg,
        sojourn_time,
        concurrent_activities,
    )


def apply_heu_pandas(
    df: pd.DataFrame, parameters: Optional[Dict[Any, Any]] = None
) -> HeuristicsNet:
    """
    Discovers an heuristics net using the Heuristics Miner ++ algorithm

    Implements the approach described in
    Burattin, Andrea, and Alessandro Sperduti. "Heuristics Miner for Time Intervals." ESANN. 2010.

    https://andrea.burattin.net/public-files/publications/2010-esann-slides.pdf

    Parameters
    --------------
    df
        Dataframe
    parameters
        Parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY
        - Parameters.START_TIMESTAMP_KEY
        - Parameters.TIMESTAMP_KEY
        - Parameters.CASE_ID_KEY
        - Parameters.DEPENDENCY_THRESH
        - Parameters.AND_MEASURE_THRESH
        - Parameters.MIN_ACT_COUNT
        - Parameters.MIN_DFG_OCCURRENCES
        - Parameters.HEU_NET_DECORATION

    Returns
    --------------
    heu_net
        Heuristics net
    """
    if parameters is None:
        parameters = {}
    (
        start_activities,
        end_activities,
        activities_occurrences,
        dfg,
        performance_dfg,
        sojourn_time,
        concurrent_activities,
    ) = discover_abstraction_dataframe(df, parameters=parameters)
    return discover_heu_net_plus_plus(
        start_activities,
        end_activities,
        activities_occurrences,
        dfg,
        performance_dfg,
        sojourn_time,
        concurrent_activities,
        parameters=parameters,
    )


def discover_abstraction_dataframe(
    df: pd.DataFrame, parameters: Optional[Dict[Any, Any]] = None
) -> Tuple[Any, Any, Any, Any, Any, Any, Any]:
    """
    Discovers an abstraction from a dataframe that is useful for the Heuristics Miner ++ algorithm

    Parameters
    --------------
    df
        Dataframe
    parameters
        Parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY
        - Parameters.START_TIMESTAMP_KEY
        - Parameters.TIMESTAMP_KEY
        - Parameters.CASE_ID_KEY

    Returns
    --------------
    start_activities
        Start activities
    end_activities
        End activities
    activities_occurrences
        Activities along with their number of occurrences
    dfg
        Directly-follows graph
    performance_dfg
        (Performance) Directly-follows graph
    sojourn_time
        Sojourn time for each activity
    concurrent_activities
        Concurrent activities
    """
    if parameters is None:
        parameters = {}
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes.DEFAULT_NAME_KEY
    )
    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY, parameters, None
    )
    if start_timestamp_key is None:
        start_timestamp_key = xes.DEFAULT_START_TIMESTAMP_KEY
        parameters = copy(parameters)
        parameters[Parameters.START_TIMESTAMP_KEY] = start_timestamp_key
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, xes.DEFAULT_TIMESTAMP_KEY
    )
    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )

    if is_polars_lazyframe(df):
        from pm4py.algo.discovery.dfg.adapters.polars import df_statistics
        from pm4py.statistics.attributes.polars import get as pd_attributes
        from pm4py.statistics.concurrent_activities.polars import get as pd_conc_act
        from pm4py.statistics.end_activities.polars import get as pd_ea
        from pm4py.statistics.eventually_follows.polars import get as pd_efg
        from pm4py.statistics.service_time.polars import get as pd_soj_time
        from pm4py.statistics.start_activities.polars import get as pd_sa
    else:
        from pm4py.algo.discovery.dfg.adapters.pandas import df_statistics
        from pm4py.statistics.attributes.pandas import get as pd_attributes
        from pm4py.statistics.concurrent_activities.pandas import get as pd_conc_act
        from pm4py.statistics.end_activities.pandas import get as pd_ea
        from pm4py.statistics.eventually_follows.pandas import get as pd_efg
        from pm4py.statistics.service_time.pandas import get as pd_soj_time
        from pm4py.statistics.start_activities.pandas import get as pd_sa

    start_activities = pd_sa.get_start_activities(df, parameters=parameters)
    end_activities = pd_ea.get_end_activities(df, parameters=parameters)
    activities_occurrences = pd_attributes.get_attribute_values(
        df, activity_key, parameters=parameters
    )
    efg_parameters = copy(parameters)
    efg_parameters[pd_efg.Parameters.KEEP_FIRST_FOLLOWING] = True
    dfg = pd_efg.apply(df, parameters=efg_parameters)
    performance_dfg = df_statistics.get_dfg_graph(
        df,
        case_id_glue=case_id_glue,
        activity_key=activity_key,
        timestamp_key=timestamp_key,
        start_timestamp_key=start_timestamp_key,
        measure="performance",
    )
    sojourn_time = pd_soj_time.apply(df, parameters=parameters)
    concurrent_activities = pd_conc_act.apply(df, parameters=parameters)
    return (
        start_activities,
        end_activities,
        activities_occurrences,
        dfg,
        performance_dfg,
        sojourn_time,
        concurrent_activities,
    )


def discover_heu_net_plus_plus(
    start_activities,
    end_activities,
    activities_occurrences,
    dfg,
    performance_dfg,
    sojourn_time,
    concurrent_activities,
    parameters: Optional[Dict[Any, Any]] = None,
):
    """
    Discovers an heuristics net using the Heuristics Miner ++ algorithm

    Implements the approach described in
    Burattin, Andrea, and Alessandro Sperduti. "Heuristics Miner for Time Intervals." ESANN. 2010.

    https://andrea.burattin.net/public-files/publications/2010-esann-slides.pdf

    Parameters
    --------------
    start_activities
        Start activities
    end_activities
        End activities
    activities_occurrences
        Activities along with their number of occurrences
    dfg
        Directly-follows graph
    performance_dfg
        (Performance) Directly-follows graph
    sojourn_time
        Sojourn time for each activity
    concurrent_activities
        Concurrent activities
    parameters
        Parameters of the algorithm, including:
        - Parameters.DEPENDENCY_THRESH
        - Parameters.AND_MEASURE_THRESH
        - Parameters.MIN_ACT_COUNT
        - Parameters.MIN_DFG_OCCURRENCES
        - Parameters.HEU_NET_DECORATION

    Returns
    --------------
    heu_net
        Heuristics net
    """
    if parameters is None:
        parameters = {}
    dependency_thresh = exec_utils.get_param_value(
        Parameters.DEPENDENCY_THRESH,
        parameters,
        defaults.DEFAULT_DEPENDENCY_THRESH,
    )
    and_measure_thresh = exec_utils.get_param_value(
        Parameters.AND_MEASURE_THRESH,
        parameters,
        defaults.DEFAULT_AND_MEASURE_THRESH,
    )
    min_act_count = exec_utils.get_param_value(
        Parameters.MIN_ACT_COUNT, parameters, defaults.DEFAULT_MIN_ACT_COUNT
    )
    min_dfg_occurrences = exec_utils.get_param_value(
        Parameters.MIN_DFG_OCCURRENCES,
        parameters,
        defaults.DEFAULT_MIN_DFG_OCCURRENCES,
    )
    heu_net_decoration = exec_utils.get_param_value(
        Parameters.HEU_NET_DECORATION, parameters, "frequency"
    )

    # filter on activity and paths occurrence
    activities_occurrences = {
        x: y for x, y in activities_occurrences.items() if y >= min_act_count
    }
    dfg = {
        x: y
        for x, y in dfg.items()
        if y >= min_dfg_occurrences
        and x[0] in activities_occurrences
        and x[1] in activities_occurrences
    }
    performance_dfg = {x: y for x, y in performance_dfg.items() if x in dfg}
    start_activities = {
        x: y
        for x, y in start_activities.items()
        if x in activities_occurrences
    }
    end_activities = {
        x: y for x, y in end_activities.items() if x in activities_occurrences
    }
    activities = list(activities_occurrences.keys())
    if heu_net_decoration == "frequency":
        heu_net = HeuristicsNet(
            dfg,
            activities=activities,
            activities_occurrences=activities_occurrences,
            start_activities=start_activities,
            end_activities=end_activities,
        )
    else:
        heu_net = HeuristicsNet(
            dfg,
            activities=activities,
            activities_occurrences=activities_occurrences,
            start_activities=start_activities,
            end_activities=end_activities,
            performance_dfg=performance_dfg,
        )
    heu_net.min_dfg_occurrences = min_dfg_occurrences
    heu_net.sojourn_times = sojourn_time
    heu_net.concurrent_activities = concurrent_activities
    return calculate(
        heu_net, dependency_thresh, and_measure_thresh, heu_net_decoration
    )


def calculate(
    heu_net: HeuristicsNet,
    dependency_thresh: float,
    and_measure_thresh: float,
    heu_net_decoration: str,
) -> HeuristicsNet:
    """
    Calculates the dependency matrix and the AND measures using the Heuristics Miner ++ formulas

    Parameters
    ----------------
    heu_net
        Heuristics net
    dependency_thresh
        Dependency threshold
    and_measure_thresh
        AND measure threshold
    heu_net_decoration
        Decoration to use (frequency/performance)

    Returns
    ----------------
    heu_net
        Heuristics net
    """
    heu_net.performance_matrix = {}
    heu_net.dependency_matrix = {}
    heu_net.dfg_matrix = {}
    for el in heu_net.dfg:
        act1 = el[0]
        act2 = el[1]
        if act1 not in heu_net.dfg_matrix:
            heu_net.dfg_matrix[act1] = {}
            heu_net.dependency_matrix[act1] = {}
            heu_net.performance_matrix[act1] = {}
        heu_net.dfg_matrix[act1][act2] = heu_net.dfg[el]
        heu_net.dependency_matrix[act1][act2] = -1
        heu_net.performance_matrix[act1][act2] = (
            heu_net.performance_dfg[el]
            if heu_net.performance_dfg and el in heu_net.performance_dfg
            else 0.0
        )
    for act1 in heu_net.activities:
        heu_net.nodes[act1] = Node(
            heu_net,
            act1,
            heu_net.activities_occurrences[act1],
            node_type=heu_net.node_type,
        )
    # calculates the dependencies between the activities
    heu_net = calculate_dependency(
        heu_net, dependency_thresh, heu_net_decoration
    )
    # calculates the AND measure for outgoing edges (e.g. which activities
    # happen in parallel after a given activity)
    heu_net = calculate_and_out_measure(heu_net, and_measure_thresh)
    # calculates the AND measure for ingoing edges (e.g. which activities
    # happen in parallel before a given activity)
    heu_net = calculate_and_in_measure(heu_net, and_measure_thresh)
    return heu_net


def calculate_dependency(
    heu_net: HeuristicsNet, dependency_thresh: float, heu_net_decoration: str
) -> HeuristicsNet:
    """
    Calculates the dependency matrix using the Heuristics Miner ++ formula

    Parameters
    --------------
    heu_net
        Heuristics net
    dependency_thresh
        Dependency threshold
    heu_net_decoration
        Decoration to include (frequency/performance)

    Returns
    ---------------
    heu_net
        Heuristics net (enriched)
    """
    for act1 in heu_net.activities:
        if act1 in heu_net.dfg_matrix:
            for act2 in heu_net.dfg_matrix[act1]:
                v1 = heu_net.dfg_matrix[act1][act2]
                v2 = (
                    heu_net.dfg_matrix[act2][act1]
                    if act2 in heu_net.dfg_matrix
                    and act1 in heu_net.dfg_matrix[act2]
                    else 0.0
                )
                tup = tuple(sorted((act1, act2)))
                # added term for Heuristics Miner ++
                v3 = (
                    heu_net.concurrent_activities[tup]
                    if tup in heu_net.concurrent_activities
                    else 0.0
                )
                dep = (v1 - v2) / (v1 + v2 + v3)
                heu_net.dependency_matrix[act1][act2] = dep
                if dep > dependency_thresh:
                    repr_value = (
                        v1
                        if heu_net_decoration == "frequency"
                        else heu_net.performance_matrix[act1][act2]
                    )
                    heu_net.nodes[act1].add_output_connection(
                        heu_net.nodes[act2], dep, v1, repr_value=repr_value
                    )
                    heu_net.nodes[act2].add_input_connection(
                        heu_net.nodes[act1], dep, v1, repr_value=repr_value
                    )
    return heu_net


def calculate_and_out_measure(
    heu_net: HeuristicsNet, and_measure_thresh: float
) -> HeuristicsNet:
    """
    Calculates the AND measure for outgoing edges using the Heuristics Miner ++ formula

    Parameters
    ---------------
    heu_net
        Heuristics net
    and_measure_thresh
        And measure threshold

    Returns
    ---------------
    heu_net
        Heuristics net (enriched)
    """
    for act in heu_net.nodes:
        nodes = sorted(
            x.node_name for x in heu_net.nodes[act].output_connections
        )
        i = 0
        while i < len(nodes):
            n1 = nodes[i]
            v3 = (
                heu_net.dfg_matrix[act][n1]
                if act in heu_net.dfg_matrix and n1 in heu_net.dfg_matrix[act]
                else 0.0
            )
            j = i + 1
            while j < len(nodes):
                n2 = nodes[j]
                tup = tuple(sorted((n1, n2)))
                v1 = (
                    heu_net.dfg_matrix[n1][n2]
                    if n1 in heu_net.dfg_matrix
                    and n2 in heu_net.dfg_matrix[n1]
                    else 0.0
                )
                v2 = (
                    heu_net.dfg_matrix[n2][n1]
                    if n2 in heu_net.dfg_matrix
                    and n1 in heu_net.dfg_matrix[n2]
                    else 0.0
                )
                v4 = (
                    heu_net.dfg_matrix[act][n2]
                    if act in heu_net.dfg_matrix
                    and n2 in heu_net.dfg_matrix[act]
                    else 0.0
                )
                # added term for Heuristics Miner ++
                v5 = (
                    heu_net.concurrent_activities[tup]
                    if tup in heu_net.concurrent_activities
                    else 0.0
                )
                this_value = (v1 + v2 + v5) / (v3 + v4)
                if this_value > and_measure_thresh:
                    if n1 not in heu_net.nodes[act].and_measures_out:
                        heu_net.nodes[act].and_measures_out[n1] = {}
                    heu_net.nodes[act].and_measures_out[n1][n2] = this_value
                j = j + 1
            i = i + 1
    return heu_net


def calculate_and_in_measure(
    heu_net: HeuristicsNet, and_measure_thresh: float
) -> HeuristicsNet:
    """
    Calculates the AND measure for incoming edges using the Heuristics Miner ++ formula

    Parameters
    ---------------
    heu_net
        Heuristics net
    and_measure_thresh
        And measure threshold

    Returns
    ---------------
    heu_net
        Heuristics net (enriched)
    """
    for act in heu_net.nodes:
        nodes = sorted(
            x.node_name for x in heu_net.nodes[act].input_connections
        )
        i = 0
        while i < len(nodes):
            n1 = nodes[i]
            v3 = (
                heu_net.dfg_matrix[n1][act]
                if n1 in heu_net.dfg_matrix and act in heu_net.dfg_matrix[n1]
                else 0.0
            )
            j = i + 1
            while j < len(nodes):
                n2 = nodes[j]
                tup = tuple(sorted((n1, n2)))
                v1 = (
                    heu_net.dfg_matrix[n1][n2]
                    if n1 in heu_net.dfg_matrix
                    and n2 in heu_net.dfg_matrix[n1]
                    else 0.0
                )
                v2 = (
                    heu_net.dfg_matrix[n2][n1]
                    if n2 in heu_net.dfg_matrix
                    and n1 in heu_net.dfg_matrix[n2]
                    else 0.0
                )
                v4 = (
                    heu_net.dfg_matrix[n2][act]
                    if n2 in heu_net.dfg_matrix
                    and act in heu_net.dfg_matrix[n2]
                    else 0.0
                )
                # added term for Heuristics Miner ++
                v5 = (
                    heu_net.concurrent_activities[tup]
                    if tup in heu_net.concurrent_activities
                    else 0.0
                )
                this_value = (v1 + v2 + v5) / (v3 + v4)
                if this_value > and_measure_thresh:
                    if n1 not in heu_net.nodes[act].and_measures_in:
                        heu_net.nodes[act].and_measures_in[n1] = {}
                    heu_net.nodes[act].and_measures_in[n1][n2] = this_value
                j = j + 1
            i = i + 1
    return heu_net


def apply_dfg(
    dfg,
    activities=None,
    activities_occurrences=None,
    start_activities=None,
    end_activities=None,
    parameters=None,
):
    raise Exception("not implemented for plusplus version")


def apply_heu_dfg(
    dfg,
    activities=None,
    activities_occurrences=None,
    start_activities=None,
    end_activities=None,
    dfg_window_2=None,
    freq_triples=None,
    parameters=None,
):
    raise Exception("not implemented for plusplus version")



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from pm4py import util
from pm4py.objects.log.obj import EventLog
from pm4py.objects.trie.obj import Trie
from pm4py.util import pandas_utils, exec_utils
from pm4py.utils import is_polars_lazyframe
from typing import Optional, Dict, Any, Union
import pandas as pd


class Parameters(Enum):
    ACTIVITY_KEY = util.constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    MAX_PATH_LENGTH = (
        "max_path_length"  # New parameter for maximum path length
    )


def apply(
    log: Union[EventLog, pd.DataFrame],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Trie:
    parameters = parameters if parameters is not None else dict()

    # Extract the maximum path length if provided
    max_path_length = exec_utils.get_param_value(
        Parameters.MAX_PATH_LENGTH, parameters, None
    )

    if pandas_utils.check_is_pandas_dataframe(log):
        if is_polars_lazyframe(log):
            from pm4py.statistics.variants.polars import get as get_variants_polars

            variants = get_variants_polars.get_variants_set(
                log, parameters=parameters
            )
        else:
            from pm4py.statistics.variants.pandas import get as get_variants_pandas

            variants = get_variants_pandas.get_variants_set(
                log, parameters=parameters
            )
    else:
        from pm4py.statistics.variants.log import get as get_variants_log

        variants = get_variants_log.get_variants(log, parameters=parameters)

    variants = list(variants)

    root = Trie()

    for variant in variants:
        # If maximum path length is defined, truncate the variant
        if max_path_length is not None and len(variant) > max_path_length:
            variant = variant[:max_path_length]

        trie = root
        for i, activity in enumerate(variant):
            match = False
            for c in trie.children:
                if c.label == activity:
                    trie = c
                    match = True
                    break
            if not match:
                node = Trie(label=activity, parent=trie, depth=trie.depth + 1)
                trie.children.append(node)
                trie = node
            # If at the end of this (possibly truncated) variant, mark as final
            if i == len(variant) - 1:
                trie.final = True

    return root



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Dict, Optional, Any, List, Union

import polars as pl

from pm4py.statistics.overlap.utils import compute
from pm4py.util import exec_utils, constants, xes_constants


class Parameters(Enum):
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY


def apply(
    lf: pl.LazyFrame,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> List[int]:
    """
    Computes the case overlap statistic from a Polars LazyFrame

    Parameters
    -----------------
    lf
        LazyFrame
    parameters
        Parameters of the algorithm, including:
        - Parameters.TIMESTAMP_KEY => attribute representing the completion timestamp
        - Parameters.START_TIMESTAMP_KEY => attribute representing the start timestamp

    Returns
    ----------------
    case_overlap
        List associating to each case the number of open cases during the life of a case
    """
    if parameters is None:
        parameters = {}

    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )

    columns = list({timestamp_key, start_timestamp_key, case_id_key})
    
    # Get min and max timestamps per case
    case_ranges = (
        lf.select(columns)
        .group_by(case_id_key)
        .agg([
            pl.col(start_timestamp_key).min().dt.timestamp().alias("min_start"),
            pl.col(timestamp_key).max().dt.timestamp().alias("max_end")
        ])
        .collect()
    )
    
    # Convert to list of tuples for the compute function
    points = []
    for row in case_ranges.iter_rows():
        points.append((row[1], row[2]))  # (min_start, max_end)

    return compute.apply(points, parameters=parameters)